import numpy

from darfix.core.zigzag_mode import reorder_frames_of_zigzag_scan
from darfix.tests.utils import createHDF5Dataset

DIM_SLOW_NPOINTS = 10
DIM_FAST_NPOINTS = 5
SCANS_COUNT = DIM_SLOW_NPOINTS * DIM_FAST_NPOINTS
FRAME_SHAPE = (2, 2)

TOLERANCE = 1e-9


def test_zigzag():

    motor_slow = numpy.linspace(1, 4, DIM_SLOW_NPOINTS).repeat(DIM_FAST_NPOINTS)
    motor_fast = numpy.tile(numpy.linspace(10, 12, DIM_FAST_NPOINTS), DIM_SLOW_NPOINTS)

    # Generate random noise in the range of [-1e-8, 1e-8]
    noise1 = numpy.random.uniform(-TOLERANCE / 2, TOLERANCE / 2, (SCANS_COUNT,))
    noise2 = numpy.random.uniform(-TOLERANCE / 2, TOLERANCE / 2, (SCANS_COUNT,))

    motor_slow = motor_slow + noise1
    motor_fast = motor_fast + noise2

    motor_fast = motor_fast.reshape((DIM_SLOW_NPOINTS, DIM_FAST_NPOINTS))
    motor_fast[1::2][:] = motor_fast[1::2, ::-1]
    motor_fast = motor_fast.flatten()

    frame = numpy.arange(1, SCANS_COUNT + 1)
    scans = frame[:, numpy.newaxis, numpy.newaxis] * numpy.ones(FRAME_SHAPE)

    dataset = createHDF5Dataset(
        scans, {"motor_fast": motor_fast, "motor_slow": motor_slow}
    )

    dataset.find_dimensions(TOLERANCE)

    assert dataset.dims[0].name == "motor_fast"
    assert dataset.dims[0].size == DIM_FAST_NPOINTS
    assert dataset.dims[1].name == "motor_slow"
    assert dataset.dims[1].size == DIM_SLOW_NPOINTS

    metadata_dict_before = dataset.metadata_dict.copy()

    reorder_frames_of_zigzag_scan(dataset.dims, dataset)

    for i in range(DIM_SLOW_NPOINTS):
        if i % 2 == 0:
            continue
        for y in range(DIM_FAST_NPOINTS):
            index = i * DIM_FAST_NPOINTS + y
            index_reverse = i * DIM_FAST_NPOINTS + (DIM_FAST_NPOINTS - 1 - y)
            numpy.testing.assert_array_equal(scans[index], dataset.data[index_reverse])
            assert (
                metadata_dict_before["motor_fast"][index]
                == dataset.metadata_dict["motor_fast"][index_reverse]
            )
            assert (
                metadata_dict_before["motor_slow"][index]
                == dataset.metadata_dict["motor_slow"][index_reverse]
            )



from __future__ import annotations

import logging

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.gui.plot.items import ImageBase
from silx.io.dictdump import dicttonx

from ... import dtypes
from ...core.grainplot import OrientationDistImage
from ...core.grainplot import generate_grain_maps_nxdict
from ..utils.utils import select_output_hdf5_file_with_dialog
from .mosaicityWidget import MosaicityWidget
from .utils import MapType
from .utils import add_image_with_transformation

_logger = logging.getLogger(__file__)


class GrainPlotWidget(qt.QMainWindow):
    """
    Widget to show a series of maps for the analysis of the data.
    """

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)

        self.dataset: dtypes.ImageDataset | None = None
        self._plots: list[Plot2D] = []

        self._mapTypeComboBox = qt.QComboBox()
        self._mapTypeComboBox.addItems(MapType.values())
        self._mapTypeComboBox.currentTextChanged.connect(self._updatePlot)

        self._plotWidget = qt.QWidget()
        plotsLayout = qt.QHBoxLayout()
        self._plotWidget.setLayout(plotsLayout)
        widget = qt.QWidget(parent=self)
        layout = qt.QVBoxLayout()

        self._mosaicity_widget = MosaicityWidget()

        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.clicked.connect(self.exportMaps)

        self._messageWidget = qt.QLabel("No dataset in input.")

        layout.addWidget(self._mapTypeComboBox)
        layout.addWidget(self._plotWidget)
        layout.addWidget(self._mosaicity_widget)
        layout.addWidget(self._messageWidget)
        layout.addWidget(self._exportButton)
        self._plotWidget.hide()
        self._mosaicity_widget.hide()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

        self._mapTypeComboBox.setDisabled(True)
        self._exportButton.setDisabled(True)

    def setMessage(self, message: str):
        self._messageWidget.setText(message)

    def setDataset(self, dataset: dtypes.Dataset):
        if len(dataset.dataset.moments_dims) == 0:
            raise ValueError(
                "Please compute moments (dataset.apply_moments) before setting the dataset"
            )

        self.dataset = dataset.dataset

        for i in reversed(range(self._plotWidget.layout().count())):
            self._plotWidget.layout().itemAt(i).widget().setParent(None)

        self._plots.clear()
        for dim in self.dataset.dims.values():
            plot = Plot2D(parent=self)
            plot.setKeepDataAspectRatio(True)
            plot.setGraphTitle(self.dataset.title + "\n" + dim.name)
            plot.setDefaultColormap(Colormap(name="viridis"))
            self._plots.append(plot)
            self._plotWidget.layout().addWidget(plot)

        # Enable mosaicity if ndim >= 2
        if self.dataset.dims.ndim >= 2:
            self._mosaicity_widget.setDataset(self.dataset)
            self._mapTypeComboBox.model().item(4).setEnabled(True)
            self._mapTypeComboBox.setCurrentText(MapType.MOSAICITY.value)
        else:
            self._mapTypeComboBox.model().item(4).setEnabled(False)
            self._mapTypeComboBox.setCurrentText(MapType.COM.value)
        # Force plot update since setCurrentText does not fire currentTextChanged if the text is the same
        # https://doc.qt.io/qt-6/qcombobox.html#currentTextChanged
        self._updatePlot(self._mapTypeComboBox.currentText())

        self._messageWidget.hide()
        self._mapTypeComboBox.setEnabled(True)
        self._exportButton.setEnabled(True)

    def _updatePlot(self, raw_map_type: str):
        """
        Update shown plots in the widget
        """

        map_type = MapType(raw_map_type)
        if map_type == MapType.MOSAICITY:
            self._mosaicity_widget.show()
            self._plotWidget.hide()
            return

        if self.dataset is None:
            return

        moments = self.dataset.moments_dims
        self._mosaicity_widget.hide()
        if map_type == MapType.FWHM:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, moments[i][1])
        elif map_type == MapType.COM:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, moments[i][0])
        elif map_type == MapType.SKEWNESS:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, moments[i][2])
        elif map_type == MapType.KURTOSIS:
            self._plotWidget.show()
            for i, plot in enumerate(self._plots):
                self._addImage(plot, moments[i][3])
        else:
            _logger.warning("Unexisting map method")

    def _generate_maps_nxdict(self) -> dict:
        orientation_image_plot: ImageBase | None = (
            self._mosaicity_widget.getContoursImage()
        )
        orientation_dist_data = self._mosaicity_widget.getOrientationDist()

        if orientation_image_plot and orientation_dist_data:
            orientation_dist_image = OrientationDistImage(
                data=orientation_dist_data.data,
                as_rgb=orientation_dist_data.as_rgb,
                origin=orientation_image_plot.getOrigin(),
                scale=orientation_image_plot.getScale(),
                xlabel=orientation_image_plot.getXLabel(),
                ylabel=orientation_image_plot.getYLabel(),
                contours=self._mosaicity_widget.contours,
            )
        else:
            orientation_dist_image = None

        return generate_grain_maps_nxdict(
            self.dataset,
            self._mosaicity_widget.getMosaicity(),
            orientation_dist_image,
        )

    def exportMaps(self):
        """
        Creates dictionary with maps information and exports it to a nexus file
        """
        nx = self._generate_maps_nxdict()

        filename = select_output_hdf5_file_with_dialog()
        if filename:
            dicttonx(nx, filename)

    def _addImage(self, plot, image):
        if self.dataset is None:
            transformation = None
        else:
            transformation = self.dataset.transformation
        add_image_with_transformation(plot, image, transformation)



from typing import Sequence
from typing import Tuple
from typing import Union

import numpy


def unique(
    ar: Sequence, return_index: bool = False, **kw
) -> Union[Tuple[numpy.ndarray], numpy.ndarray]:
    """Returns the unique elements of an array while preserving their order.

    The signature of this function is the same as `numpy.unique`.
    """
    unique_values, unique_indices, *other_indices = numpy.unique(
        ar, return_index=True, **kw
    )

    idx = numpy.argsort(unique_indices)

    unique_values = unique_values[idx]
    if not return_index and not other_indices:
        return unique_values
    unique_indices = unique_indices[idx]
    if not other_indices:
        return unique_values, unique_indices
    other_indices = tuple(values[idx] for values in other_indices)
    if not return_index:
        return (unique_values,) + other_indices
    return (unique_values, unique_indices) + other_indices



import importlib.metadata
import logging
import sys
from datetime import datetime

import h5py
import numpy
from silx.io.dictdump import dicttoh5
from silx.io.dictdump import h5todict

_logger = logging.getLogger(__file__)

_VERSION = importlib.metadata.version("darfix")


def assert_string(s, enums):
    s += " has to be "
    for app in enums:
        if app == enums[0]:
            s += "`" + app + "`"
        elif app == enums[-1]:
            s += "or `" + app + "`"
        else:
            s += ", `" + app + "`"
    return s


# Print iterations progress
def advancement_display(
    ncurrent,
    ntotal,
    prefix="",
    suffix="",
    decimals=1,
    length=100,
    fill="\u2588",
    left="-",
):
    """
    Call in a loop to create terminal progress bar
    @params:
        ncurrent   - Required  : current N (Int)
        ntotal     - Required  : total expected N (Int)
        prefix     - Optional  : prefix string (Str)
        suffix     - Optional  : suffix string (Str)
        decimals   - Optional  : positive number of decimals in percent complete (Int)
        length     - Optional  : character length of bar (Int)
        fill       - Optional  : bar fill character (Str)
        left       - Optional  : bar fill character (Str)
    """
    if ntotal:
        ncurrent = min(ncurrent, ntotal)
        progress = ncurrent / float(ntotal)
    else:
        progress = ncurrent + 1
    percent_fmt = "{0:." + str(decimals) + "f}"
    percent = percent_fmt.format(100 * progress)
    filledLength = int(length * min(progress, 1))
    leftLength = length - filledLength
    bar = fill * filledLength + left * leftLength
    sys.stdout.write("\r%s |%s| %s%% %s\r" % (prefix, bar, percent, suffix))
    # Print New Line on Complete
    if ncurrent >= ntotal:
        sys.stdout.write("\n")


def write_maps(
    h5_file,
    list_of_maps,
    default_map,
    entry,
    processing_order,
    data_path="/",
    overwrite=True,
):
    """
    Write a stack of components and its parameters into .h5

    :param str h5_file: path to the hdf5 file
    :param str entry: entry name
    :param dict dimensions: Dictionary with the dimensions names and values
    :param numpy.ndarray W: Matrix with the rocking curves values
    :param numpy.ndarray data: Stack with the components
    :param int processing_order: processing order of treatment
    :param str data_path: path to store the data
    """
    process_name = "process_" + str(processing_order)

    def get_interpretation(my_data):
        """Return hdf5 attribute for this type of data"""
        if isinstance(my_data, numpy.ndarray):
            if my_data.ndim == 1:
                return "spectrum"
            elif my_data.ndim in (2, 3):
                return "image"
        return None

    def save_key(path_name, key_path, value, overwrite=True):
        """Save the given value to the associated path. Manage numpy arrays
        and dictionaries"""
        key_path = key_path.replace(".", "/")
        # save if is dict
        if isinstance(value, dict):
            h5_path = "/".join((path_name, key_path))
            dicttoh5(
                value, h5file=h5_file, h5path=h5_path, overwrite_data=True, mode="a"
            )
        else:
            with h5py.File(h5_file, "a") as h5f:
                nx = h5f.require_group(path_name)
                if overwrite and key_path in nx:
                    del nx[key_path]
                try:
                    nx[key_path] = value
                except TypeError as e:
                    _logger.error("Unable to write", str(key_path), "reason is", str(e))
                else:
                    interpretation = get_interpretation(value)
                    if interpretation:
                        nx[key_path].attrs["interpretation"] = interpretation

    with h5py.File(h5_file, "a") as h5f:
        h5f.attrs["default"] = entry
        nx_entry = h5f.require_group("/".join((data_path, entry)))
        nx_entry.attrs["NX_class"] = "NXentry"
        nx_entry.attrs["default"] = "data"

        nx_process = nx_entry.require_group(process_name)
        nx_process.attrs["NX_class"] = "NXprocess"
        if overwrite:
            for key in ("program", "version", "date", "processing_order"):
                if key in nx_process:
                    del nx_process[key]
        nx_process["program"] = "darfix"
        nx_process["version"] = _VERSION
        nx_process["date"] = datetime.now().replace(microsecond=0).isoformat()
        nx_process["processing_order"] = numpy.int32(processing_order)

        results = nx_process.require_group("results")
        results.attrs["NX_class"] = "NXcollection"
        nx_data = nx_entry.require_group("data")
        nx_data.attrs["NX_class"] = "NXdata"
        default = list_of_maps[default_map]
        source_addr = entry + "/" + process_name + "/results/" + default_map
        results.attrs["target"] = default_map
        save_key(results.name, default_map, default)
        save_key(nx_data.name, default_map, h5f[source_addr])

        for _map in list_of_maps:
            if _map == default_map:
                continue
            if isinstance(list_of_maps[_map], dict):
                maps = results.require_group(_map)
                maps.attrs["NX_class"] = "NXcollection"
                for method in list_of_maps[_map]:
                    save_key(maps.name, method, list_of_maps[_map][method])
            else:
                save_key(results.name, _map, list_of_maps[_map])


def read_components(h5_file):
    """
    Read a stack of components and its parameters from a Nexus file.

    :param str h5_file: path to the hdf5 file
    """
    with h5py.File(h5_file, "r") as nx:
        # find the default NXentry group
        nx_entry = nx[nx.attrs["default"]]
        # find the default NXdata group
        nx_process = nx_entry["process_1"]
        input_data = nx_process["dimensions"]
        dimensions = h5todict(
            h5_file, nx.attrs["default"] + "/process_1/dimensions", asarray=False
        )
        input_data = nx_process["values"]
        values = {}
        for key in input_data.keys():
            values[key] = numpy.array(list(input_data[key]))
        results = nx_process["results"]
        components = numpy.array(list(results["components"]))
        W = numpy.array(list(results["W"]))

    return dimensions, components, W, values


def write_components(
    h5_file,
    entry,
    dimensions,
    W,
    data,
    values,
    processing_order,
    data_path="/",
    overwrite=True,
):
    """
    Write a stack of components and its parameters into .h5

    :param str h5_file: path to the hdf5 file
    :param str entry: entry name
    :param dict dimensions: Dictionary with the dimensions names and values
    :param numpy.ndarray W: Matrix with the rocking curves values
    :param numpy.ndarray data: Stack with the components
    :param int processing_order: processing order of treatment
    :param str data_path: path to store the data
    """
    process_name = "process_" + str(processing_order)

    def get_interpretation(my_data):
        """Return hdf5 attribute for this type of data"""
        if isinstance(my_data, numpy.ndarray):
            if my_data.ndim == 1:
                return "spectrum"
            elif my_data.ndim in (2, 3):
                return "image"
        return None

    def save_key(path_name, key_path, value, overwrite=True):
        """Save the given value to the associated path. Manage numpy arrays
        and dictionaries"""
        key_path = key_path.replace(".", "/")
        # save if is dict
        if isinstance(value, dict):
            h5_path = "/".join((path_name, key_path))
            dicttoh5(
                value, h5file=h5_file, h5path=h5_path, overwrite_data=True, mode="a"
            )
        else:
            with h5py.File(h5_file, "a") as h5f:
                nx = h5f.require_group(path_name)
                if overwrite and key_path in nx:
                    del nx[key_path]
                try:
                    nx[key_path] = value
                except TypeError as e:
                    _logger.error("Unable to write", str(key_path), "reason is", str(e))
                else:
                    interpretation = get_interpretation(value)
                    if interpretation:
                        nx[key_path].attrs["interpretation"] = interpretation

    with h5py.File(h5_file, "a") as h5f:
        h5f.attrs["default"] = entry
        nx_entry = h5f.require_group("/".join((data_path, entry)))
        nx_entry.attrs["NX_class"] = "NXentry"
        nx_entry.attrs["default"] = "data"

        nx_process = nx_entry.require_group(process_name)
        nx_process.attrs["NX_class"] = "NXprocess"
        if overwrite:
            for key in ("program", "version", "date", "processing_order"):
                if key in nx_process:
                    del nx_process[key]
        nx_process["program"] = "darfix"
        nx_process["version"] = _VERSION
        nx_process["date"] = datetime.now().replace(microsecond=0).isoformat()
        nx_process["processing_order"] = numpy.int32(processing_order)

        nx_parameters = nx_process.require_group("dimensions")
        nx_parameters.attrs["NX_class"] = "NXparameters"
        dicttoh5(dimensions, h5f, entry + "/" + process_name + "/dimensions")

        nx_values = nx_process.require_group("values")
        nx_values.attrs["NX_class"] = "NXparameters"
        for key, value in values.items():
            save_key(nx_values.name, key_path=key, value=value)

        results = nx_process.require_group("results")
        results.attrs["NX_class"] = "NXcollection"

        nx_data = nx_entry.require_group("data")
        nx_data.attrs["NX_class"] = "NXdata"
        nx_data.attrs["signal"] = "components"
        source_addr = entry + "/" + process_name + "/results/components"
        results.attrs["target"] = "components"

        save_key(results.name, "W", W)
        save_key(results.name, "components", data)
        save_key(nx_data.name, "components", h5f[source_addr])


def create_nxdata_dict(
    signal: numpy.ndarray,
    signal_name: str,
    axes=None,
    axes_names=None,
    axes_long_names=None,
    rgba=False,
):
    nxdata = {signal_name: signal, "@signal": signal_name, "@NX_class": "NXdata"}

    if axes is not None:
        nxdata.update(
            {"@axes": axes_names, axes_names[0]: axes[0], axes_names[1]: axes[1]}
        )

        if axes_long_names is not None:
            nxdata.update(
                {
                    axes_names[0] + "@long_name": axes_long_names[0],
                    axes_names[1] + "@long_name": axes_long_names[1],
                }
            )

    if rgba:
        nxdata[signal_name + "@interpretation"] = "rgba-image"
        nxdata[signal_name + "@CLASS"] = "IMAGE"

    return nxdata



from __future__ import annotations

from typing import Literal

from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from ..dtypes import Dataset
from ..pixel_sizes import PixelSize


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """Input dataset containing a stack of images."""
    kind: Literal["rsm", "magnification"] | bool | MissingData = MISSING_DATA
    "Kind of transformation to apply. RSM only applicable for 1D datasets. 'magnification' is used if kind is not provided.",
    orientation: int | MissingData = MISSING_DATA
    """Used only with kind='magnification'."""
    magnification: float | MissingData = MISSING_DATA
    """To be used only with kind='magnification'. Magnification factor to apply to the dataset."""
    pixelSize: Literal["Basler", "PcoEdge_2x", "PcoEdge_10x"] | MissingData = (
        MISSING_DATA
    )
    """To be used only with kind='rsm', distance in micrometers of each pixel."""
    rotate: bool | MissingData = MISSING_DATA
    """To be used only with kind='rsm', if True the images with transformation are rotated 90 degrees."""


class TransformationMatrixComputation(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """Computes transformation matrix and attach it to the dataset"""

    def run(self):
        input_dataset: Dataset = self.inputs.dataset
        if not isinstance(input_dataset, Dataset):
            raise TypeError(
                f"Dataset is expected to be an instance of {Dataset}. Got {type(input_dataset)} instead."
            )
        dataset = input_dataset.dataset

        if not dataset.dims.ndim:
            return input_dataset

        kind = self.get_input_value("kind", None)

        # Dirty workaround to handle both bool and str for kind because of legacy 2.X
        if isinstance(kind, bool):
            kind = "rsm" if kind else "magnification"

        if kind == "rsm":
            assert dataset.dims.ndim == 1, "Kind RSM can only be used for 1D datasets."

            pixelSize: str = self.get_input_value("pixelSize", None)
            rotate: bool = self.get_input_value("rotate", None)

            dataset.compute_transformation(
                PixelSize[pixelSize].value, kind=kind, rotate=rotate
            )
        else:
            magnification: float = self.get_input_value("magnification", None)
            orientation: int = self.get_input_value("orientation", None)

            if orientation == -1 or orientation is None:
                dataset.compute_transformation(magnification, kind=kind)
            else:
                dataset.compute_transformation(
                    magnification, topography_orientation=orientation
                )

        self.outputs.dataset = Dataset(
            dataset=dataset,
            indices=input_dataset.indices,
            bg_indices=input_dataset.bg_indices,
            bg_dataset=input_dataset.bg_dataset,
        )



from __future__ import annotations

import numpy
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix.dtypes import Dataset
from darfix.gui.rsmHistogramWidget import RSMHistogramWidget
from darfix.tasks.rsm_histogram import RSMHistogram


class RSMHistogramWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=RSMHistogram):
    """
    Widget to compute Reciprocal Space Map
    """

    name = "rsm histogram"
    icon = "icons/category.svg"
    want_main_area = False

    _ewoks_inputs_to_hide_from_orange = (
        "Q",
        "a",
        "map_range",
        "detector",
        "units",
        "n",
        "map_shape",
        "energy",
    )

    def __init__(self):
        super().__init__()

        self._widget = RSMHistogramWidget(parent=self)
        self._widget.sigComputeClicked.connect(self._compute)
        self.controlArea.layout().addWidget(self._widget)
        q = self.get_default_input_value("Q")
        if q:
            self._widget.q = q
        a = self.get_default_input_value("a")
        if a:
            self._widget.a = a
        map_range = self.get_default_input_value("map_range")
        if map_range:
            self._widget.map_range = map_range
        detector = self.get_default_input_value("detector")
        if detector:
            self._widget.detector = detector
        units = self.get_default_input_value("units")
        if units:
            self._widget.units = units
        n = self.get_default_input_value("n")
        if n:
            self._widget.n = n
        map_shape = self.get_default_input_value("map_shape")
        if map_shape:
            self._widget.map_shape = map_shape
        energy = self.get_default_input_value("energy")
        if energy:
            self._widget.energy = energy

    def handleNewSignals(self):
        dataset: Dataset | None = self.get_task_input_value("dataset", None)
        if dataset is None:
            return
        self.setDataset(dataset)

        # Do not call super().handleNewSignals() to prevent propagation

    def setDataset(self, dataset: Dataset):
        self._widget.setDataset(dataset)
        self.open()

    def _compute(self):
        self.set_dynamic_input("dataset", self._widget.dataset)
        self.set_default_input("Q", self._widget.q.tolist())
        self.set_default_input("a", self._widget.a)
        self.set_default_input("map_range", self._widget.map_range)
        self.set_default_input("detector", self._widget.detector)
        self.set_default_input("units", self._widget.units)
        self.set_default_input("n", self._widget.n.tolist())
        self.set_default_input("map_shape", self._widget.map_shape.tolist())
        self.set_default_input("energy", self._widget.energy)

        self.execute_ewoks_task_without_propagation()

    def task_output_changed(self) -> None:
        arr: numpy.ndarray | None = self.get_task_output_value("hist_values", None)
        edges: numpy.ndarray | None = self.get_task_output_value("hist_edges", None)

        if arr is None or edges is None:
            return
        self._widget.updatePlot(arr, edges)



import math

import numpy

TWO_PI = 2 * numpy.pi
SQRT_2 = math.sqrt(2)


class OperationAborted(Exception):
    """Raised when operation is aborted"""

    def __init__(self) -> None:
        super().__init__("Operation aborted.")


class NoDimensionsError(Exception):
    """Error raised when a method needing Darfix dimensions is called before the dimensions were found."""

    def __init__(self, method_name: str) -> None:
        super().__init__(
            f"{method_name} needs to have defined dimensions. Run `find_dimensions` before `{method_name}`."
        )


class TooManyDimensionsForRockingCurvesError(ValueError):
    def __init__(self) -> None:
        super().__init__(
            "Unsupported number of dimensions. Rocking curves only support 1D, 2D or 3D datasets."
        )


def wrapTo2pi(x):
    """
    Python implementation of Matlab method `wrapTo2pi`.
    Wraps angles in x, in radians, to the interval [0, 2*pi] such that 0 maps
    to 0 and 2*pi maps to 2*pi. In general, positive multiples of 2*pi map to
    2*pi and negative multiples of 2*pi map to 0.
    """
    xwrap = numpy.remainder(x - numpy.pi, TWO_PI)
    mask = numpy.abs(xwrap) > numpy.pi
    xwrap[mask] -= TWO_PI * numpy.sign(xwrap[mask])
    return xwrap + numpy.pi


def compute_hsv(x_data: numpy.ndarray, y_data: numpy.ndarray):
    data = numpy.arctan2(-y_data, -x_data)

    hue = wrapTo2pi(data) / TWO_PI
    saturation = numpy.sqrt(numpy.power(x_data, 2) + numpy.power(y_data, 2)) / SQRT_2
    value = numpy.ones_like(x_data)

    # Display NaN values as black
    value[numpy.isnan(data)] = 0
    hue[numpy.isnan(data)] = 0
    saturation[numpy.isnan(data)] = 0

    return numpy.stack(
        (
            hue,
            saturation,
            value,
        ),
        axis=2,
    )



from __future__ import annotations

import logging
import os

import silx.io
from ewoksorange.gui.qtsignals import block_signals
from silx.gui import qt

from darfix.gui.data_selection.utils import find_detector_name
from darfix.gui.data_selection.utils import find_scan_names
from darfix.gui.utils.message import show_error_msg

from .line_edits import DatasetLineEdit
from .line_edits import FileLineEdit
from .line_edits import GroupLineEdit

_logger = logging.getLogger(__file__)


class ScanSelectionWidget(qt.QWidget):
    """
    Widget to select a scan (file + scan_number) and a detector in the scan.

    Autofills the detector path when the scan is updated.
    """

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent=parent, **kwargs)

        self._fileLineEdit = FileLineEdit()
        self._scanNameComboBox = qt.QComboBox()
        self._detectorLineEdit = DatasetLineEdit()

        layout = qt.QFormLayout(self)

        layout.addRow("File selection", self._fileLineEdit)
        layout.addRow("Scan number", self._scanNameComboBox)
        layout.addRow("Detector", self._detectorLineEdit)

        self._scanNameComboBox.currentTextChanged.connect(self._updateDetectorPath)
        self._fileLineEdit.editingFinished.connect(self._onNewFileEdit)
        self._fileLineEdit.dialogSelected.connect(self._onNewFile)

    def _onNewFileEdit(self):
        path = self._fileLineEdit.getText()
        if os.path.exists(path):
            self._onNewFile(path)
        else:
            show_error_msg(f"File '{path}' does not exist")

    def _onNewFile(self, filePath: str):
        self._detectorLineEdit.setFile(filePath)

        try:
            scanNames = find_scan_names(filePath)
        except (FileNotFoundError, PermissionError, IsADirectoryError) as e:
            _logger.warning(f"Could not find scan names in {filePath}. Reason: {e}")
            scanNames = tuple()

        with block_signals(self._scanNameComboBox):
            self._scanNameComboBox.clear()
        self._scanNameComboBox.addItems(scanNames)

    def _updateDetectorPath(self, scanNumber: str):
        scanUrl = f"{self._fileLineEdit.getText()}::/{scanNumber}"
        with silx.io.open(scanUrl) as scan:
            detectorName = find_detector_name(scan)

        if detectorName:
            self._detectorLineEdit.setText(f"/{scanNumber}/measurement/{detectorName}")

    def setFilePath(self, file: str):
        self._fileLineEdit.setText(file)
        self._onNewFile(file)

    def getFilePath(self) -> str:
        return self._fileLineEdit.getText()

    def setDetectorPath(self, path: str):
        self._detectorLineEdit.setText(path)
        scanNumber = path.lstrip("/").split("/")[0]
        with block_signals(self._scanNameComboBox):
            self._scanNameComboBox.setCurrentText(scanNumber)

    def getDetectorPath(self) -> str:
        return self._detectorLineEdit.getText()


class ScanSelectionWidgetWithPositioners(ScanSelectionWidget):
    """
    Adds an autofillable positioner path to ScanSelectionWidget
    """

    def __init__(self, parent: qt.QWidget | None = None, **kwargs) -> None:
        super().__init__(parent, **kwargs)
        layout = self.layout()
        assert isinstance(layout, qt.QFormLayout)

        self._positionersLineEdit = GroupLineEdit()
        layout.addRow("Positioners", self._positionersLineEdit)

        self._scanNameComboBox.currentTextChanged.connect(self._updatePositionersPath)

    def _onNewFile(self, file_path: str):
        super()._onNewFile(file_path)
        self._positionersLineEdit.setFile(file_path)

    def _updatePositionersPath(self, scan_number: str):
        self._positionersLineEdit.setText(f"/{scan_number}/instrument/positioners")

    def setPositionersPath(self, data_url: str):
        self._positionersLineEdit.setText(data_url)

    def getPositionersPath(self) -> str:
        return self._positionersLineEdit.getText()



from __future__ import annotations

from typing import Any

import numpy
from attr import dataclass
from matplotlib.colors import hsv_to_rgb
from silx.math.combo import min_max
from silx.utils.enum import Enum

from darfix.io.utils import create_nxdata_dict
from darfix.math import sturges_rules

from ..dtypes import AxisType
from .dataset import ImageDataset
from .utils import compute_hsv


class MomentType(Enum):
    COM = "Center of mass"
    FWHM = "FWHM"
    SKEWNESS = "Skewness"
    KURTOSIS = "Kurtosis"


class OrientationDistData:
    KEY_SCALE = 1000

    def __init__(
        self,
        dataset: ImageDataset,
        x_dimension: int,
        y_dimension: int,
    ) -> None:

        # automatic bins
        x_bins = sturges_rules(
            dataset.dims.get(x_dimension).size * numpy.prod(dataset.frame_shape)
        )
        y_bins = sturges_rules(
            dataset.dims.get(y_dimension).size * numpy.prod(dataset.frame_shape)
        )

        self.data, self.x_range, self.y_range = (
            compute_orientation_distribution_histogram(
                dataset, x_dimension, y_dimension, x_bins, y_bins
            )
        )

        self.x_bins = x_bins
        self.y_bins = y_bins
        self.x_label = dataset.dims.get(x_dimension).name
        self.y_label = dataset.dims.get(y_dimension).name

        x_data = numpy.linspace(-1, 1, self.KEY_SCALE)
        y_data = numpy.linspace(-1, 1, self.KEY_SCALE)
        x_mesh, y_mesh = numpy.meshgrid(x_data, y_data)
        self.rgb_key = hsv_to_rgb(compute_hsv(x_mesh, y_mesh))

    def origin(
        self,
        origin: AxisType,
    ) -> tuple[float, float]:
        if origin == "dims":
            return (self.x_range[0], self.y_range[0])
        elif origin == "center":
            return (
                -numpy.ptp(self.x_range) / 2,
                -numpy.ptp(self.y_range) / 2,
            )
        else:
            return (0, 0)

    def data_plot_scale(self) -> tuple[float, float]:
        return (
            numpy.ptp(self.x_range) / self.x_bins,
            numpy.ptp(self.y_range) / self.y_bins,
        )

    def rgb_key_plot_scale(self) -> tuple[float, float]:
        return (
            numpy.ptp(self.x_range) / self.KEY_SCALE,
            numpy.ptp(self.y_range) / self.KEY_SCALE,
        )

    def to_motor_coordinates(
        self,
        points_x: numpy.ndarray,
        points_y: numpy.ndarray,
        origin: AxisType,
    ) -> tuple[numpy.ndarray, numpy.ndarray]:
        """
        Given points_x, points_y in the 2D space of self.data, returns motor coordinates x, y
        """
        x_origin, y_origin = self.origin(origin)
        return (
            points_x * numpy.ptp(self.x_range) / (self.x_bins - 1) + x_origin,
            points_y * numpy.ptp(self.y_range) / (self.y_bins - 1) + y_origin,
        )


@dataclass
class OrientationDistImage:
    xlabel: str
    ylabel: str
    scale: tuple[float, float]
    origin: tuple[float, float]
    data: numpy.ndarray
    as_rgb: numpy.ndarray
    contours: dict


class MultiDimMomentType(Enum):
    """Moments that are only computed for datasets with multiple dimensions"""

    ORIENTATION_DIST = "Orientation distribution"
    MOSAICITY = "Mosaicity"


def get_axes(transformation: numpy.ndarray | None) -> tuple[
    tuple[numpy.ndarray, numpy.ndarray] | None,
    tuple[str, str] | None,
    tuple[str, str] | None,
]:
    if not transformation:
        return None, None, None

    axes = (transformation.yregular, transformation.xregular)
    axes_names = ("y", "x")
    axes_long_names = (transformation.label, transformation.label)

    return axes, axes_names, axes_long_names


def compute_normalized_component(component: numpy.ndarray):

    min_max_result = min_max(component)
    min_component = min_max_result.minimum
    max_component = min_max_result.maximum

    return 2 * (component - min_component) / (max_component - min_component) - 1


def compute_mosaicity(
    moments: dict[int, numpy.ndarray], x_dimension: int, y_dimension: int
):
    norm_center_of_mass_x = compute_normalized_component(moments[x_dimension][0])
    norm_center_of_mass_y = compute_normalized_component(moments[y_dimension][0])

    return hsv_to_rgb(compute_hsv(norm_center_of_mass_x, norm_center_of_mass_y))


def create_moment_nxdata_groups(
    parent: dict[str, Any],
    moment_data: numpy.ndarray,
    axes,
    axes_names,
    axes_long_names,
):

    for i, map_type in enumerate(MomentType.values()):
        parent[map_type] = create_nxdata_dict(
            moment_data[i],
            map_type,
            axes,
            axes_names,
            axes_long_names,
        )


def generate_grain_maps_nxdict(
    dataset: ImageDataset,
    mosaicity: numpy.ndarray | None,
    orientation_dist_image: OrientationDistImage | None,
) -> dict:
    moments = dataset.moments_dims
    axes, axes_names, axes_long_names = get_axes(dataset.transformation)

    nx = {
        "entry": {"@NX_class": "NXentry"},
        "@NX_class": "NXroot",
        "@default": "entry",
    }

    if mosaicity is not None:
        nx["entry"][MultiDimMomentType.MOSAICITY.value] = create_nxdata_dict(
            mosaicity,
            MultiDimMomentType.MOSAICITY.value,
            axes,
            axes_names,
            axes_long_names,
            rgba=True,
        )
        nx["entry"]["@default"] = MultiDimMomentType.MOSAICITY.value
    else:
        nx["entry"]["@default"] = MomentType.COM.value

    if orientation_dist_image is not None:
        nx["entry"][MultiDimMomentType.ORIENTATION_DIST.value] = {
            "key": {
                "image": orientation_dist_image.as_rgb,
                "data": orientation_dist_image.data,
                "origin": orientation_dist_image.origin,
                "scale": orientation_dist_image.scale,
                "xlabel": orientation_dist_image.xlabel,
                "ylabel": orientation_dist_image.ylabel,
                "image@interpretation": "rgba-image",
                "image@CLASS": "IMAGE",
            },
            "curves": orientation_dist_image.contours,
        }

    if dataset.dims.ndim <= 1:
        create_moment_nxdata_groups(
            nx["entry"],
            moments[0],
            axes,
            axes_names,
            axes_long_names,
        )
    else:
        for axis, dim in dataset.dims.items():
            nx["entry"][dim.name] = {"@NX_class": "NXcollection"}
            create_moment_nxdata_groups(
                nx["entry"][dim.name],
                moments[axis],
                axes,
                axes_names,
                axes_long_names,
            )

    return nx


def compute_orientation_distribution_histogram(
    dataset: ImageDataset,
    x_dimension: int,
    y_dimension: int,
    x_bins: int,
    y_bins: int,
) -> tuple[numpy.ndarray, tuple[float, float], tuple[float, float]]:
    """
    Computes the orientation distribution hostogram.

    Requires apply_moments to be called before.

    :returns: 2d histogram, x range, y range.

    """

    if len(dataset.moments_dims) == 0:
        raise ValueError("Moments should be computed before to use this function.")

    com_x = dataset.moments_dims[x_dimension][0].ravel()
    com_y = dataset.moments_dims[y_dimension][0].ravel()

    range_x = [numpy.nanmin(com_x), numpy.nanmax(com_x)]
    range_y = [numpy.nanmin(com_y), numpy.nanmax(com_y)]

    # Histogram in 2D
    orientation_distribution, _, _ = numpy.histogram2d(
        com_y,
        com_x,  # note: y first is in purpose : see numpy.histogram2d documentation
        weights=dataset.zsum().ravel(),  # We need to take into account pixel intensity
        bins=[y_bins, x_bins],
        range=[range_y, range_x],
    )

    return orientation_distribution, range_x, range_y



from __future__ import annotations

import os

import numpy
from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from pydantic import Field
from silx.utils.enum import Enum as _Enum

from darfix import dtypes
from darfix.io.utils import write_components


class Method(_Enum):
    """
    Different blind source separation approaches that can be applied
    """

    PCA = "PCA"

    NICA = "NICA"

    NMF = "NMF"

    NICA_NMF = "NICA_NMF"

    @classmethod
    def from_value(cls, value):
        # ensure backward compatibility with ewoks <= 1.0
        # the value of the Method use to be the description / tooltip
        for method, desc in Method._descriptions().items():
            if value == desc:
                return method
        return super().from_value(value)

    @staticmethod
    def _descriptions() -> dict:
        return {
            Method.PCA: (
                "The process of computing the principal components \n"
                "and using them to perform a change of basis on the data"
            ),
            Method.NICA: "Find components independent from each other and non-negative",
            Method.NMF: (
                "Non-negative matrix factorization factorizes the data matrix into \n"
                "two matrices, with the property that all three matrices have no negative elements"
            ),
            Method.NICA_NMF: "Apply Non-negative ICA followed by NMF",
        }

    @staticmethod
    def get_description(method) -> str:
        method = Method.from_value(method)
        if method in Method._descriptions():
            return Method._descriptions()[method]
        else:
            raise NotImplementedError


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """ Input dataset containing a stack of images """
    method: Method
    "Method to use for blind source separation"
    n_comp: int | MissingData = Field(
        default=MISSING_DATA, description="Number of components to extract"
    )
    save: bool | MissingData = MISSING_DATA
    processing_order: int | MissingData = MISSING_DATA


class BlindSourceSeparation(
    Task,
    input_model=Inputs,
    output_names=["dataset", "comp", "W"],
):
    """Perform blind source separation on a Darfix dataset.
    Blind source separation (BSS) comprises all techniques that try to decouple a set of source signals from a set of mixed signals with unknown (or very little) information.

    Supported methods are PCA, NICA, NMF and NICA_NMF.

    Related article  : https://pmc.ncbi.nlm.nih.gov/articles/PMC10161887/#sec3.3.3

    """

    def run(self):
        if not isinstance(self.inputs.dataset, dtypes.Dataset):
            raise TypeError(
                f"'dataset' input should be an instance of Dataset. Got {type(self.inputs.dataset)}."
            )
        dataset = self.inputs.dataset.dataset
        indices = self.inputs.dataset.indices
        bg_indices = self.inputs.dataset.bg_indices
        bg_dataset = self.inputs.dataset.bg_dataset

        n_comp = self.get_input_value("n_comp", None)
        method = Method.from_value(self.inputs.method)
        if method == Method.PCA:
            comp, W = dataset.pca(n_comp, indices=indices)
        elif method == Method.NICA:
            comp, W = dataset.nica(n_comp, indices=indices)
        elif method == Method.NMF:
            comp, W = dataset.nmf(n_comp, indices=indices)
        elif method == Method.NICA_NMF:
            comp, W = dataset.nica_nmf(n_comp, indices=indices)
        else:
            raise ValueError("BSS method not managed")
        n_comp = comp.shape[0]
        shape = dataset.frame_shape
        comp = comp.reshape(n_comp, shape[0], shape[1])
        if bg_indices is not None:
            # If filter data is activated, the matrix W has reduced dimensionality, so reshaping is not possible
            # Create empty array with shape the total number of frames
            W = numpy.zeros((dataset.nframes, n_comp))
            # Set actual values of W where threshold of filter is True
            W[indices] = W
            W = W

        if self.get_input_value("save", False):
            write_components(
                h5file=os.path.join(dataset.dir, "components.h5"),
                entry="entry",
                dimensions=dataset.dims.to_dict(),
                W=W,
                data=comp,
                values=dataset.get_dimensions_values(),
                processing_order=self.get_input_value("processing_order", 0),
            )
        self.outputs.dataset = dtypes.Dataset(
            dataset=dataset,
            indices=indices,
            bg_indices=bg_indices,
            bg_dataset=bg_dataset,
        )

        self.outputs.W = W
        self.outputs.comp = comp



from __future__ import annotations

from typing import List

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot import Plot2D
from silx.io.dictdump import dicttonx

from ..dtypes import Dataset
from ..io.utils import create_nxdata_dict
from ..pixel_sizes import PixelSize
from .utils.utils import select_output_hdf5_file_with_dialog


class _LineEditsWidget(qt.QWidget):
    def __init__(self, parent, dims=1, validator=None, placeholder=None):
        qt.QWidget.__init__(self, parent)
        layout = qt.QHBoxLayout()
        if placeholder:
            assert len(placeholder) == dims
        self._lineEdits = []
        for i in range(dims):
            lineEdit = qt.QLineEdit(parent=self)
            if placeholder:
                lineEdit.setPlaceholderText(placeholder[i])
            if validator:
                lineEdit.setValidator(validator)
            self._lineEdits.append(lineEdit)
            layout.addWidget(lineEdit)

        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    @property
    def value(self):
        values = []
        for le in self._lineEdits:
            values += [le.text() if le.text() != "" else None]
        return values

    def setValue(self, values: List[int]):
        assert type(values) is list

        for i, value in enumerate(values):
            self._lineEdits[i].setText(str(value))
            self._lineEdits[i].setCursorPosition(0)


class RSMHistogramWidget(qt.QWidget):
    """
    Widget to compute Reciprocal Space Map
    """

    sigComputeClicked = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.dataset: Dataset | None = None

        self._idx = [(2, 1), (2, 0), (1, 0)]

        layout = qt.QGridLayout()
        label = qt.QLabel("Q:")
        font = label.font()
        font.setBold(True)
        label.setFont(font)
        layout.addWidget(label, 0, 0, 1, 1)
        self._q = _LineEditsWidget(self, 3, qt.QIntValidator(), ["h", "k", "l"])
        self._q.setValue([1, 0, 1])
        layout.addWidget(self._q, 0, 1, 1, 2)
        label = qt.QLabel("Pseudocubic lattice:")
        label.setFont(font)
        layout.addWidget(label, 1, 0, 1, 2)
        self._a = qt.QLineEdit()
        self._a.setValidator(qt.QDoubleValidator())
        self.a = 4.08
        layout.addWidget(self._a, 1, 1, 1, 2)
        label = qt.QLabel("Map range:")
        label.setFont(font)
        layout.addWidget(label, 2, 0, 1, 2)
        self._map_range = qt.QLineEdit()
        self._map_range.setValidator(qt.QDoubleValidator())
        self.map_range = 0.008
        layout.addWidget(self._map_range, 2, 1, 1, 2)
        label = qt.QLabel("Detector:")
        label.setFont(font)
        layout.addWidget(label, 3, 0, 1, 2)
        self._detector = qt.QComboBox()
        self._detector.addItems(PixelSize.names())
        layout.addWidget(self._detector, 3, 1, 1, 2)
        label = qt.QLabel("Units:")
        label.setFont(font)
        layout.addWidget(label, 4, 0, 1, 2)
        self._units = qt.QComboBox()
        self._units.addItems(["Poulsen", "Gorfman"])
        layout.addWidget(self._units, 4, 1, 1, 2)
        label = qt.QLabel("n:")
        label.setFont(font)
        layout.addWidget(label, 5, 0, 1, 1)
        self._n = _LineEditsWidget(self, 3, qt.QIntValidator(), ["h", "k", "l"])
        self._n.setValue([0, 1, 0])
        layout.addWidget(self._n, 5, 1, 1, 2)
        label = qt.QLabel("Map shape:")
        label.setFont(font)
        layout.addWidget(label, 6, 0, 1, 1)
        self._map_shape = _LineEditsWidget(self, 3, qt.QIntValidator(), ["x", "y", "z"])
        self._map_shape.setValue([200, 200, 200])
        layout.addWidget(self._map_shape, 6, 1, 1, 2)
        label = qt.QLabel("Energy:")
        label.setFont(font)
        layout.addWidget(label, 7, 0, 1, 2)
        self._energy = qt.QLineEdit()
        self._energy.setValidator(qt.QDoubleValidator())
        self.energy = 17
        layout.addWidget(self._energy, 7, 1, 1, 2)
        self._computeB = qt.QPushButton("Compute")
        layout.addWidget(self._computeB, 8, 2, 1, 1)
        self._computeB.clicked.connect(self.sigComputeClicked.emit)
        self._plotWidget = qt.QWidget()
        self._plotWidget.setLayout(qt.QHBoxLayout())
        layout.addWidget(self._plotWidget, 9, 0, 1, 3)
        self._plotWidget.hide()
        self._exportButton = qt.QPushButton("Export maps")
        self._exportButton.hide()
        layout.addWidget(self._exportButton, 10, 2, 1, 1)
        self._exportButton.clicked.connect(self.exportMaps)

        self.setLayout(layout)
        self.setMinimumWidth(650)

    def setDataset(self, dataset: Dataset):
        self.dataset = dataset
        for i in reversed(range(self._plotWidget.layout().count())):
            self._plotWidget.layout().itemAt(i).widget().setParent(None)

        self._plotWidget.hide()
        self._exportButton.hide()

        self._plots = []
        for i in range(len(self._idx)):
            self._plots += [Plot2D(parent=self)]
            self._plots[-1].setDefaultColormap(Colormap(name="viridis"))
            self._plotWidget.layout().addWidget(self._plots[-1])

    def _toFloat(self, text: str) -> float:
        value, ok = self.locale().toFloat(text)
        if not ok:
            raise ValueError(f"Could not convert string to float: '{text}'")
        return value

    def _toStr(self, value: float) -> str:
        return self.locale().toString(value)

    @property
    def q(self):
        return numpy.array(self._q.value, dtype=int)

    @q.setter
    def q(self, q: List[int]):
        self._q.setValue(q)

    @property
    def a(self):
        return self._toFloat(self._a.text())

    @a.setter
    def a(self, a: float):
        self._a.setText(self._toStr(a))

    @property
    def map_range(self):
        return self._toFloat(self._map_range.text())

    @map_range.setter
    def map_range(self, map_range: float):
        self._map_range.setText(self._toStr(map_range))

    @property
    def detector(self):
        return self._detector.currentText()

    @detector.setter
    def detector(self, detector: str):
        self._detector.setCurrentText(detector)

    @property
    def units(self):
        return self._units.currentText()

    @units.setter
    def units(self, units: str):
        self._units.setCurrentText(units)

    @property
    def n(self):
        return numpy.array(self._n.value, dtype=int)

    @n.setter
    def n(self, n: List[int]):
        self._n.setValue(n)

    @property
    def map_shape(self):
        return numpy.array(self._map_shape.value, dtype=int)

    @map_shape.setter
    def map_shape(self, map_shape: List[int]):
        self._map_shape.setValue(map_shape)

    @property
    def energy(self):
        return self._toFloat(self._energy.text())

    @energy.setter
    def energy(self, energy: float):
        self._energy.setText(self._toStr(energy))

    def updatePlot(self, arr: numpy.ndarray, edges: numpy.ndarray):
        arr = numpy.nan_to_num(arr)

        self._projections = []

        if self._units.currentText().lower() == "poulsen":
            self.labels = [r"$q_{rock}$", r"$q_{\perp}$", r"$q_{||}$"]
        else:
            self.labels = ["h", "k", "l"]

        for idx, (i, j) in enumerate(self._idx):
            self._projections += [numpy.sum(arr, axis=idx)]
            xscale = (edges[i][-1] - edges[i][0]) / len(edges[i])
            yscale = (edges[j][-1] - edges[j][0]) / len(edges[j])
            self._plots[idx].addImage(
                self._projections[-1],
                scale=(xscale, yscale),
                origin=(edges[i][0], edges[j][0]),
                xlabel=self.labels[i],
                ylabel=self.labels[j],
            )

        self._edges = edges

        self._plotWidget.show()
        self._exportButton.show()

    def exportMaps(self):
        """
        Creates dictionary with maps information and exports it to a nexus file
        """
        entry = "entry"

        nx = {
            entry: {"@NX_class": "NXentry"},
            "@NX_class": "NXroot",
            "@default": "entry",
        }

        for idx, (j, i) in enumerate(self._idx):
            axis = [self._edges[i][:-1], self._edges[j][:-1]]
            nx["entry"][str(idx)] = create_nxdata_dict(
                self._projections[idx],
                self.labels[i] + "_" + self.labels[j],
                axis,
                axes_names=[self.labels[i], self.labels[j]],
            )

        filename = select_output_hdf5_file_with_dialog()
        if filename:
            dicttonx(nx, filename)



import os

import pytest
from ewoksorange.tests.conftest import qtapp  # noqa F401

from darfix.core.datapathfinder import DETECTOR_KEYWORD
from darfix.core.datapathfinder import FIRST_SCAN_KEYWORD
from darfix.core.datapathfinder import LAST_SCAN_KEYWORD
from darfix.core.datapathfinder import SCAN_KEYWORD
from darfix.gui.utils.data_path_completer import DataPathLineEditWithCompleter
from darfix.gui.utils.data_path_completer import DataPathModel
from darfix.tests.utils import create_scans


@pytest.mark.parametrize("display_dataset", (True, False))
def test_DataPathModel_scan_and_detector_keywords(
    tmp_path, qtapp, display_dataset: bool  # noqa F811
):
    """
    test keywords handling in the case we have '{scan}' and '{detector}' keywords
    """
    test_folder = tmp_path / "test_DataPathModel_scan_and_detector_keywords"
    test_folder.mkdir()
    test_file = os.path.join(test_folder, "scans")
    create_scans(file_path=test_file, n_scan=2)

    allowed_keywords = (SCAN_KEYWORD, DETECTOR_KEYWORD)
    widget = DataPathLineEditWithCompleter(
        allowed_keywords=allowed_keywords, completer_display_dataset=display_dataset
    )
    widget.setFile(test_file)

    completer_model = widget.completer().model()
    assert isinstance(completer_model, DataPathModel)

    assert completer_model.getAllowedKeywords() == allowed_keywords

    # check that the 'root' keywords are at root
    assert completer_model.rowCount() == 1
    assert completer_model.columnCount() == 1
    assert len(completer_model.findItems(SCAN_KEYWORD)) == 1
    assert len(completer_model.findItems(FIRST_SCAN_KEYWORD)) == 0
    assert len(completer_model.findItems(LAST_SCAN_KEYWORD)) == 0
    # check the 'scan' keyword tree is correctly solved
    scan_item = completer_model.findItems(SCAN_KEYWORD)[0]
    assert scan_item.columnCount() == 1
    assert scan_item.rowCount() == 3
    scan_root_det = scan_item.child(0, 0)
    scan_instrument = scan_item.child(1, 0)
    assert scan_root_det.text() == DETECTOR_KEYWORD
    assert scan_instrument.text() == "instrument"
    measurement_item = scan_item.child(2, 0)
    assert measurement_item.text() == "measurement"

    # make sure the r'{detector}' keyword is a leaf
    assert scan_root_det.rowCount() == 0
    assert scan_root_det.columnCount() == 0

    # check that the 2.1/instrument item is correctly solved until the end
    assert scan_instrument.rowCount() == 2
    assert scan_instrument.columnCount() == 1
    detector_item = scan_instrument.child(1, 0)
    assert detector_item.text() == DETECTOR_KEYWORD
    measurement_item = scan_instrument.child(0, 0)

    positioners_item = scan_instrument.child(0, 0)
    assert positioners_item.text() == "positioners"

    assert positioners_item.rowCount() == (4 if display_dataset else 0)
    assert positioners_item.columnCount() == (1 if display_dataset else 0)


@pytest.mark.parametrize("display_dataset", (True, False))
def test_DataPathModel_first_scan_and_last_scan_and_detectorkeywords(
    tmp_path, qtapp, display_dataset: bool  # noqa F811
):
    """
    test keywords handling in the case we have '{first_scan}', '{last_scan}' and {detector} keywords
    """
    test_folder = tmp_path / "test_DataPathModel_first_scan_and_last_scan_keywords"
    test_folder.mkdir()
    test_file = os.path.join(test_folder, "scans")
    create_scans(file_path=test_file, n_scan=2)

    allowed_keywords = (FIRST_SCAN_KEYWORD, LAST_SCAN_KEYWORD, DETECTOR_KEYWORD)
    widget = DataPathLineEditWithCompleter(
        allowed_keywords=allowed_keywords, completer_display_dataset=display_dataset
    )
    widget.setFile(test_file)
    completer_model = widget.completer().model()
    assert isinstance(completer_model, DataPathModel)

    assert completer_model.getAllowedKeywords() == allowed_keywords

    # check that the 'root' keywords are at root
    assert completer_model.rowCount() == 4  # {first_scan}, {last_scan}, 1.1, 2.1
    assert completer_model.columnCount() == 1
    assert len(completer_model.findItems("")) == 1  # file root level
    assert len(completer_model.findItems(SCAN_KEYWORD)) == 0
    assert len(completer_model.findItems(FIRST_SCAN_KEYWORD)) == 1
    assert len(completer_model.findItems(LAST_SCAN_KEYWORD)) == 1
    # check the 'scan' keyword tree is correctly solved
    scan_item = completer_model.findItems(FIRST_SCAN_KEYWORD)[0]
    assert scan_item.columnCount() == 1
    assert scan_item.rowCount() == 3
    scan_root_det = scan_item.child(0, 0)
    scan_instrument = scan_item.child(1, 0)
    assert scan_root_det.text() == DETECTOR_KEYWORD
    assert scan_instrument.text() == "instrument"
    measurement_item = scan_item.child(2, 0)
    assert measurement_item.text() == "measurement"

    # make sure the r'{detector}' keyword is a leaf
    assert scan_root_det.rowCount() == 0
    assert scan_root_det.columnCount() == 0

    # check that the 2.1/instrument item is correctly solved until the end
    assert scan_instrument.rowCount() == 2
    assert scan_instrument.columnCount() == 1
    detector_item = scan_instrument.child(1, 0)
    assert detector_item.text() == DETECTOR_KEYWORD
    measurement_item = scan_instrument.child(0, 0)

    positioners_item = scan_instrument.child(0, 0)
    assert positioners_item.text() == "positioners"

    assert positioners_item.rowCount() == (4 if display_dataset else 0)
    assert positioners_item.columnCount() == (1 if display_dataset else 0)

    # check one real scan path
    existing_path_item = completer_model.findItems("")[0]
    assert existing_path_item.rowCount() == 3  # 1.1, 2.1 and {detector}
    assert existing_path_item.columnCount() == 1
    item_1_1 = existing_path_item.child(0, 0)
    assert item_1_1.text() == "1.1"
    item_2_1 = existing_path_item.child(2, 0)
    assert item_2_1.text() == "2.1"
    item_detector = existing_path_item.child(1, 0)
    assert item_detector.text() == DETECTOR_KEYWORD

    assert item_1_1.rowCount() == 3  # instrument, measurement and {detector}



from __future__ import annotations

import logging

import numpy
from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from darfix.core.dataset import ImageDataset
from darfix.dtypes import Dataset

logger = logging.getLogger(__file__)


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """Input dataset containing a stack of images."""
    roi_origin: numpy.ndarray | list[float] | MissingData = MISSING_DATA
    """Origin of the ROI to apply. If not provided, dataset will be unchanged."""
    roi_size: numpy.ndarray | list[float] | MissingData = MISSING_DATA
    """Size of the ROI to apply. If not provided, dataset will be unchanged."""


class RoiSelection(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """Apply a Region of Interest (ROI) selection on a Darfix dataset."""

    def run(self):
        input_dataset: Dataset = self.inputs.dataset
        dataset: ImageDataset = input_dataset.dataset
        bg_dataset: ImageDataset | None = input_dataset.bg_dataset

        origin: numpy.ndarray = numpy.flip(self.get_input_value("roi_origin", []))
        size: numpy.ndarray = numpy.flip(self.get_input_value("roi_size", []))

        frame_shape: numpy.ndarray = numpy.flip(dataset.frame_shape)

        if len(origin) == 0 or len(size) == 0:
            # ROI undefined
            logger.warning(
                f"Cannot apply a ROI if origin ({origin}) or size ({size}) is empty. Dataset is unchanged."
            )
        elif tuple(origin) == (0, 0) and tuple(size) == tuple(frame_shape):
            logger.info("ROI is covering the full frame. Dataset is unchanged.")
            # ROI covering the frame shape
            pass
        else:
            dataset = dataset.apply_roi(origin=origin, size=size)
            if bg_dataset:
                bg_dataset = bg_dataset.apply_roi(origin=origin, size=size)

        self.outputs.dataset = Dataset(
            dataset=dataset,
            indices=input_dataset.indices,
            bg_indices=input_dataset.bg_indices,
            bg_dataset=bg_dataset,
        )



from __future__ import annotations

from silx.gui import qt

from ...core.noiseremoval import NoiseRemovalOperation
from ...core.noiseremoval import operation_to_str


class OperationHistoryWidget(qt.QWidget):
    """Keeps the history of noise removal operations and displays them in a QListWidget"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        layout = qt.QVBoxLayout()
        assert layout is not None
        self._listWidget = qt.QListWidget()
        self._listWidget.setSelectionMode(
            qt.QAbstractItemView.SelectionMode.NoSelection
        )
        self._listWidget.hide()
        self._checkbox = qt.QCheckBox("Show history")

        layout.addWidget(self._checkbox)
        layout.addWidget(self._listWidget)
        self.setLayout(layout)

        self._checkbox.toggled.connect(self._listWidget.setVisible)
        self._checkbox.setChecked(True)

        self._operations: list[NoiseRemovalOperation] = []

    def append(self, operation: NoiseRemovalOperation):
        self._operations.append(operation)
        operation_order = len(self._operations)
        self._listWidget.addItem(
            qt.QListWidgetItem(f"{operation_order}: {str(operation)}")
        )

    def pop(self):
        self._operations.pop()
        self._listWidget.takeItem(self._listWidget.count() - 1)

    def clear(self):
        self._operations.clear()
        self._listWidget.clear()

    def getOperations(self) -> list[NoiseRemovalOperation]:
        return list(self._operations)

    def extend(self, operations: list[NoiseRemovalOperation]):
        existing_count = len(self._operations)
        self._operations.extend(operations)
        for i_new_op, operation in enumerate(operations):
            self._listWidget.addItem(
                qt.QListWidgetItem(
                    f"{existing_count + i_new_op}: {operation_to_str(operation)}"
                )
            )



import os
import urllib.parse
import urllib.request
from typing import Optional
from typing import Union

from silx.io.url import DataUrl

from darfix.core.dataset import ImageDataset


def load_process_data(
    detector_url: Union[str, DataUrl],
    root_dir: Optional[str] = None,
    dark_detector_url: Optional[Union[str, DataUrl]] = None,
    title: str = "",
    metadata_url=None,
):
    """
    Loads data from `detector_url`.
    If `detector_url` is:

        - a str: consider it as a file pattern (for EDF files).
        - a DataUrl: consider it readable by silx `get_data` function

    :param detector_url: detector_url to be loaded.
    :param metadata_url: path to the scan metadata for HDF5 containing positioner information in order to load metadata for non-edf files
    """
    indices = li_indices = None
    root_dir_specified = bool(root_dir)

    if isinstance(detector_url, DataUrl):
        assert detector_url.file_path() not in (
            "",
            None,
        ), "no file_path provided to the DataUrl"
        if not root_dir_specified:
            root_dir = os.path.dirname(detector_url.file_path())
        dataset = ImageDataset(
            _dir=root_dir,
            detector_url=detector_url,
            title=title,
            metadata_url=metadata_url,
        )
    elif isinstance(detector_url, str):
        if not detector_url:
            raise ValueError("'detector_url' cannot be an empty string")
        if not root_dir_specified:
            root_dir = _get_root_dir(detector_url)
        dataset = ImageDataset(
            _dir=root_dir,
            detector_url=detector_url,
            title=title,
            metadata_url=metadata_url,
        )
    else:
        raise TypeError(
            f"Expected detector_url to be a string or a silx DataUrl. Got {type(detector_url)} instead."
        )

    if not dark_detector_url:
        bg_dataset = None
    elif isinstance(dark_detector_url, str):
        dark_root_dir = os.path.join(dataset.dir, "dark")
        os.makedirs(dark_root_dir, exist_ok=True)
        bg_dataset = ImageDataset(
            _dir=dark_root_dir,
            detector_url=dark_detector_url,
            metadata_url=None,
        )
    elif isinstance(dark_detector_url, DataUrl):
        assert dark_detector_url.file_path() not in (
            "",
            None,
        ), "no file_path provided to the DataUrl"
        dark_root_dir = os.path.join(dataset.dir, "dark")
        os.makedirs(dark_root_dir, exist_ok=True)
        bg_dataset = ImageDataset(
            _dir=dark_root_dir,
            detector_url=dark_detector_url,
            metadata_url=None,
        )
    else:
        raise TypeError(
            f"Expected dark_detector_url to be a string or a silx DataUrl. Got {type(dark_detector_url)} instead."
        )

    assert dataset.data is not None and dataset.data.size > 0, "No data was loaded!"

    return dataset, indices, li_indices, bg_dataset


def _get_root_dir(filename: str) -> str:
    url = urllib.parse.urlparse(filename, scheme="file")
    return os.path.dirname(urllib.request.url2pathname(url.path))



from __future__ import annotations

from contextlib import contextmanager
from enum import IntEnum
from typing import Optional


class Operation(IntEnum):
    """
    Flags for different operations in Dataset
    """

    PARTITION = 0
    BS = 1
    HP = 2
    THRESHOLD = 3
    SHIFT = 4
    ROI = 5
    MOMENTS = 6
    FIT = 7
    BINNING = 8
    MASK = 9


class StateOfOperations:
    def __init__(self):
        self._is_running = [False] * len(Operation)

    def _start(self, operation: Optional[Operation]) -> None:
        if operation is None:
            return
        self._is_running[operation] = True

    def stop(self, operation: Optional[Operation]) -> None:
        if operation is None:
            return
        self._is_running[operation] = False

    @contextmanager
    def run_context(self, operation: Optional[Operation]):
        self._start(operation)
        try:
            yield
        finally:
            self.stop(operation)

    def is_running(self, operation: Operation) -> bool:
        return self._is_running[operation]



import numpy
import pytest
from scipy.stats import special_ortho_group

from darfix.decomposition.ipca import IPCA


@pytest.fixture
def rstate():
    return numpy.random.RandomState(seed=1000)


def test_singular_values(rstate):
    data = rstate.random((100, 1000))
    ipca = IPCA(data, 50, num_components=3)
    assert ipca.singular_values is None
    ipca.fit_transform()
    assert len(ipca.singular_values) == 3


def test_fit_transform(rstate):
    # Number of features (dimension of the sample space)
    n = 10
    # Number of samples
    k = 100000
    # Number of principal components (dimension of the affine subspace)
    r = 5

    # Build principal components
    all_dims = numpy.arange(n)
    rstate.shuffle(all_dims)
    sub_dims = all_dims[:r]
    V = numpy.eye(n)[:, sub_dims]  # Subset of the standard basis vectors
    R = special_ortho_group.rvs(n)  # Random rotation in R^n
    V = numpy.dot(R, V)  # Principal components
    mean = rstate.rand(n)  # Affine space translation

    # Build observations
    D = numpy.diag(range(r, 0, -1)) ** 2  # Matrix of decreasing eigenvalues
    Z = rstate.normal(size=(k, r))  # Dimensionality-reduced samples
    Z = numpy.dot(Z, D)
    X = numpy.dot(Z, V.T) + mean  # Observations

    # Apply PCA to X to get Z, V
    ipca = IPCA(X, chunksize=10000, num_components=r)
    ipca.fit_transform()
    estimated_Z = numpy.round(ipca.W, 1)
    estimated_V = numpy.round(ipca.H.T, 1)

    numpy.testing.assert_allclose(numpy.abs(estimated_Z), numpy.abs(Z), rtol=0, atol=1)
    numpy.testing.assert_allclose(numpy.abs(estimated_V), numpy.abs(V), rtol=0, atol=1)


def test_fit_transform_indices(rstate):
    # Number of features (dimension of the sample space)
    n = 10
    # Number of samples
    k = 100000
    # Indices
    indices = numpy.arange(50000)
    # Number of principal components (dimension of the affine subspace)
    r = 5

    # Build principal components
    all_dims = numpy.arange(n)
    rstate.shuffle(all_dims)
    sub_dims = all_dims[:r]
    V = numpy.eye(n)[:, sub_dims]  # Subset of the standard basis vectors
    R = special_ortho_group.rvs(n)  # Random rotation in R^n
    V = numpy.dot(R, V)  # Principal components
    mean = rstate.rand(n)  # Affine space translation

    # Build observations
    D = numpy.diag(range(r, 0, -1)) ** 2  # Matrix of decreasing eigenvalues
    Z = rstate.normal(size=(k, r))  # Dimensionality-reduced samples
    Z = numpy.dot(Z, D)
    X = numpy.dot(Z, V.T) + mean  # Observations

    # Apply PCA to X to get Z, V
    ipca = IPCA(X, chunksize=10000, num_components=r, indices=indices)
    ipca.fit_transform()
    estimated_Z = numpy.round(ipca.W, 1)
    estimated_V = numpy.round(ipca.H.T, 1)

    numpy.testing.assert_allclose(
        numpy.abs(estimated_Z), numpy.abs(Z[indices]), rtol=0, atol=1
    )
    numpy.testing.assert_allclose(numpy.abs(estimated_V), numpy.abs(V), rtol=0, atol=1)



from __future__ import annotations

from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix import dtypes
from darfix.gui.metadataWidget import MetadataWidget
from darfix.tasks.metadata import MetadataTask


class MetadataWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=MetadataTask):
    """Widget used to show the metadata in a table."""

    name = "metadata"
    icon = "icons/metadata.svg"
    want_control_area = False
    want_main_area = True

    def __init__(self):
        super().__init__()

        self._widget = MetadataWidget()
        self.mainArea.layout().addWidget(self._widget)

    def setDataset(self, dataset: dtypes.Dataset | None):
        if dataset is None:
            self._widget.clearTable()
        else:
            self._widget.setDataset(dataset)

    def handleNewSignals(self) -> None:
        dataset = self.get_task_input_value("dataset", None)
        if dataset is None:
            return
        self.setDataset(dataset)
        # This is a task only displaying metadata. there is no real processing.



from __future__ import annotations

from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import is_missing_data
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread
from ewoksorange.gui.parameterform import block_signals

from darfix import dtypes
from darfix.gui.binningWidget import BinningWidget
from darfix.tasks.binning import Binning


class BinningWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=Binning):
    """
    Widget that computes dataset binning
    """

    name = "binning"
    icon = "icons/resize.png"
    want_main_area = True

    _ewoks_inputs_to_hide_from_orange = ("scale", "output_dir")

    def __init__(self):
        super().__init__()

        self._widget = BinningWidget(parent=self)
        self.mainArea.layout().addWidget(self._widget)

        # connect signal / slot
        self._widget.sigScaleChanged.connect(self._scaleChanged)
        self._widget.sigApply.connect(self._apply)
        self._widget.sigAbort.connect(self.cancel_running_task)
        self._widget.sigComputed.connect(self.propagate_downstream)

        # set up
        scale = self.get_default_input_value("scale", MISSING_DATA)
        if scale not in (None, MISSING_DATA):
            with block_signals(self._widget):
                self._widget.scale = scale
        else:
            self._scaleChanged(self._widget.scale)

    def _scaleChanged(self, scale: float):
        self.set_default_input("scale", scale)

    def setDataset(
        self,
        dataset: dtypes.Dataset | None,
        pop_up=False,
    ):
        if dataset is None or is_missing_data(dataset):
            return

        if not isinstance(dataset, dtypes.Dataset):
            raise dtypes.DatasetTypeError(dataset)

        self._widget.setDataset(dataset)

        if pop_up:
            # raise the QDialog
            self.open()

    def handleNewSignals(self) -> None:
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.

        note: this widget can receive one signal: 'dataset'. The 'dataset' signal is handled by the ewoks task.
              This function will be only triggered when the 'dataset' signal is send
        """
        # update GUI from received dataset
        dataset = self.get_task_input_value("dataset")
        self.setDataset(dataset=dataset, pop_up=True)
        # avoid computing and propagation: waiting for apply to be pressed for now
        # super().handleNewSignals()

    def task_output_changed(self):
        # once the output is computed we can update the GUI (calling widget._displayComponents)
        dataset = self.get_task_output_value("dataset")
        self._widget._endComputation()
        if dataset not in (None, MISSING_DATA):
            self._widget.updateResultDataset(dataset.dataset)
        super().task_output_changed()

    def _apply(self):
        # in this case we only want to have the output of the task and avoid propagating until user press 'ok'
        self.execute_ewoks_task_without_propagation()



from __future__ import annotations

from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict

from darfix import dtypes


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: dtypes.Dataset
    """ Input dataset containing a stack of images """
    bins: int | MissingData = MISSING_DATA
    """ Number of bins to use for partitioning the data. Default is the number of frames in the dataset."""
    filter_bottom_bin_idx: int | MissingData = MISSING_DATA
    """ index of the bins to retrieve bottom threshold filter value. If not defined, no filtering is applied."""
    filter_top_bin_idx: int | MissingData = MISSING_DATA
    """ index of the bins to retrieve top threshold filter value. If not defined, no filtering is applied."""


class DataPartition(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """
    Filter frames with low intensity.
    """

    def run(self):
        dataset = self.inputs.dataset
        if not isinstance(dataset, dtypes.Dataset):
            raise TypeError(
                f"dataset is expected to be an instance of {dtypes.Dataset}. Got {type(dataset)}"
            )

        darfix_dataset = dataset.dataset

        indices, bg_indices = darfix_dataset.partition_by_intensity(
            bins=self.get_input_value("bins", None),
            bottom_bin=self.get_input_value("filter_bottom_bin_idx", None),
            top_bin=self.get_input_value("filter_top_bin_idx", None),
        )
        self.outputs.dataset = dtypes.Dataset(
            dataset=darfix_dataset,
            indices=indices,
            bg_indices=bg_indices,
            bg_dataset=dataset.bg_dataset,
        )



import numpy
import pytest

from darfix.core.utils import NoDimensionsError

from .utils import create_2_motors_metadata
from .utils import createDataset

N_FRAMES_DIM0 = 10
N_FRAMES_DIM1 = 5


@pytest.fixture
def dataset(tmpdir):
    n_frames = N_FRAMES_DIM0 * N_FRAMES_DIM1
    dims = (n_frames, 100, 100)
    data = numpy.zeros(dims)

    for i in range(n_frames):
        data[i] = i

    dataset = createDataset(
        data=data,
        metadata_dict=create_2_motors_metadata(N_FRAMES_DIM1, N_FRAMES_DIM0),
        _dir=str(tmpdir),
    )
    return dataset


def test_project_data(dataset):
    data = dataset.as_array3d()

    with pytest.raises(NoDimensionsError):
        proj_data = dataset.project_data(dimension=[0])

    dataset.find_dimensions()
    dataset.reshape_data()
    assert len(dataset.dims) == 2

    proj_dataset = dataset.project_data(dimension=[0])
    assert len(proj_dataset.dims) == 1
    assert proj_dataset.dims[0] == dataset.dims[0]
    proj_data = proj_dataset.as_array3d()
    for i in range(N_FRAMES_DIM0):
        numpy.testing.assert_allclose(proj_data[i], data[i::N_FRAMES_DIM0].sum(axis=0))

    proj_dataset = dataset.project_data(dimension=[1])
    assert len(proj_dataset.dims) == 1
    assert proj_dataset.dims[0] == dataset.dims[1]
    proj_data = proj_dataset.as_array3d()
    for i in range(N_FRAMES_DIM1):
        numpy.testing.assert_allclose(
            proj_data[i],
            data[i * N_FRAMES_DIM0 : (i + 1) * N_FRAMES_DIM0].sum(axis=0),
        )



import numpy
import pytest

try:
    import scipy
except ImportError:
    pytest.skip("scipy is missing", allow_module_level=True)

from darfix.core import roi
from darfix.processing import imageOperations
from darfix.tests import utils


@pytest.fixture
def dataset(tmpdir):
    first_frame = numpy.zeros((100, 100))
    # Simulating a series of frame with information in the middle.
    first_frame[25:75, 25:75] = numpy.random.randint(50, 300, size=(50, 50))
    data = [first_frame]
    shift = [1.0, 0]
    for i in range(9):
        data += [
            numpy.fft.ifftn(
                scipy.ndimage.fourier_shift(numpy.fft.fftn(data[-1]), shift)
            ).real
        ]
    data = numpy.asanyarray(data, dtype=numpy.int16)
    return utils.createDataset(data=data, _dir=str(tmpdir))


@pytest.fixture
def dark_dataset(tmpdir):
    background = [
        numpy.random.randint(-5, 5, size=(100, 100), dtype="int16") for i in range(5)
    ]
    return utils.createDataset(data=background, _dir=str(tmpdir))


def test_workflow0(dataset, dark_dataset):
    """Tests a possible workflow"""

    expected = numpy.subtract(
        dataset.data[:, 49:52, 50:51],
        numpy.median(dark_dataset.data[:, 49:52, 50:51], axis=0).astype(numpy.int16),
        dtype=numpy.int64,
    ).astype(numpy.int16)
    expected[expected > 20] = 0
    expected[expected < 1] = 0
    # ROI of the data
    data = roi.apply_3D_ROI(
        dataset.data,
        size=[3, 3],
        center=(numpy.array(dataset.data[0].shape) / 2).astype(int),
    )

    # ROI of the dark frames
    dark_frames = roi.apply_3D_ROI(
        numpy.array(dark_dataset.data),
        size=[3, 3],
        center=(numpy.array(dark_dataset.data[0].shape) / 2).astype(int),
    )

    # Background substraction of the data
    imageOperations.background_subtraction(
        data, imageOperations.compute_background(dark_frames, "median")
    )
    # ROI of the data
    data = roi.apply_3D_ROI(
        data, size=[3, 1], center=(numpy.array(data[0].shape) / 2).astype(int)
    )
    # Threshold removal of the data
    imageOperations.threshold_removal(data, 1, 20)
    numpy.testing.assert_array_equal(data, expected)


def test_workflow1(dataset, dark_dataset):
    """Tests a possible workflow"""

    first_frame = numpy.asarray(dataset.data[0])
    expected = (
        numpy.tile(first_frame, (10, 1))
        .reshape(10, 100, 100)[:, 25:75, 25:75]
        .astype(numpy.float32)
    ).copy()
    # Detect and apply the shift
    dataset.find_and_apply_shift()
    # ROI of the data
    dataset = dataset.apply_roi(
        size=[50, 50],
        center=(numpy.array(dataset.data[0].shape) / 2).astype(int),
    )

    numpy.testing.assert_allclose(
        dataset.data[0, 0:10, 0:10], expected[0, 0:10, 0:10], rtol=0.1
    )



from __future__ import annotations

from typing import Tuple

from silx.gui import qt

from ..utils.vspacer import VSpacer


class ShiftInput(qt.QWidget):
    """
    Widget used to obtain the double parameters for the shift correction.
    """

    shiftChanged = qt.Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.findShiftB = qt.QPushButton("Find shift")
        self.abortShiftB = qt.QPushButton("Abort")
        # First dim is displayed on vertical axis, second on horizontal
        firstDimLabel = qt.QLabel("Vertical shift per frame (pixels)")
        secondDimLabel = qt.QLabel("Horizontal shift per frame (pixels)")
        self._firstDimLE = qt.QLineEdit("0.0")
        self._secondDimLE = qt.QLineEdit("0.0")
        self.correctionB = qt.QPushButton("Correct")

        self._firstDimLE.setValidator(qt.QDoubleValidator())
        self._secondDimLE.setValidator(qt.QDoubleValidator())

        self._firstDimLE.editingFinished.connect(self.shiftChanged.emit)
        self._secondDimLE.editingFinished.connect(self.shiftChanged.emit)

        layout = qt.QGridLayout()

        layout.addWidget(self.findShiftB, 0, 0, 1, 1)
        layout.addWidget(self.abortShiftB, 0, 1, 1, 1)
        layout.addWidget(firstDimLabel, 1, 0)
        layout.addWidget(secondDimLabel, 2, 0)
        layout.addWidget(self._firstDimLE, 1, 1)
        layout.addWidget(self._secondDimLE, 2, 1)
        layout.addWidget(self.correctionB, 4, 0, 1, 2)

        layout.addWidget(VSpacer())

        self.setLayout(layout)

    def getShift(self) -> Tuple[float, float]:
        return float(self._firstDimLE.text()), float(self._secondDimLE.text())

    def setShift(self, shift: Tuple[float, float]):
        first_dim_shift, second_dim_shift = shift
        self._firstDimLE.setText(str(first_dim_shift))
        self._firstDimLE.setCursorPosition(0)
        self._secondDimLE.setText(str(second_dim_shift))
        self._secondDimLE.setCursorPosition(0)



from __future__ import annotations

from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

import darfix

from ..dtypes import Dataset
from .filterByDimension import FilterByDimensionWidget


class ZSumWidget(qt.QMainWindow):
    sigAxisChanged = qt.Signal(int)
    sigResetFiltering = qt.Signal()

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)

        self._sv = StackViewMainWindow()
        self._sv.setColormap(
            Colormap(
                name=darfix.config.DEFAULT_COLORMAP_NAME,
                normalization=darfix.config.DEFAULT_COLORMAP_NORM,
            )
        )
        self._sv.setKeepDataAspectRatio(True)

        self._selector = FilterByDimensionWidget()

        self._selector.axisChanged.connect(self.sigAxisChanged)
        self._selector.stateDisabled.connect(self.sigResetFiltering)

        self.setCentralWidget(self._sv)
        toolbar = qt.QToolBar()
        toolbar.addWidget(self._selector)
        self.addToolBar(qt.Qt.ToolBarArea.BottomToolBarArea, toolbar)

    def setDataset(self, dataset: Dataset):
        self.dataset = dataset.dataset
        self.indices = dataset.indices
        self._selector.setDimensions(self.dataset.dims)
        self._sv.setGraphTitle(self.dataset.title)

    def setZSum(self, zsum):
        self._sv.setStack(zsum)



import numpy

from ..dtypes import Dataset
from ..pixel_sizes import PixelSize
from ..tasks.rsm_histogram import RSMHistogram
from .utils import create_dataset_for_RSM


def test_rsm_histogram(tmpdir):
    dataset = create_dataset_for_RSM(tmpdir)
    dataset.find_dimensions()
    dataset.compute_transformation(PixelSize["Basler"].value, kind="rsm")

    arr, edges = dataset.compute_rsm(
        Q=(1, 0, 1),
        a=4.08,
        map_range=0.008,
        pixel_size=0.051,
        units="poulsen",
        n=(0, 1, 0),
        map_shape=(200, 200, 200),
    )
    assert isinstance(arr, numpy.ndarray)
    assert len(edges) > 0
    assert isinstance(edges[0], numpy.ndarray)


def test_ewoks_task(tmpdir):
    dataset = create_dataset_for_RSM(tmpdir)
    dataset.find_dimensions()
    dataset.compute_transformation(PixelSize["Basler"].value, kind="rsm")

    task = RSMHistogram(
        inputs=dict(
            dataset=Dataset(dataset),
            Q=(1, 0, 1),
            a=4.08,
            map_range=0.008,
            detector="Basler",
        )
    )
    task.execute()
    values = task.get_output_value("hist_values")
    assert isinstance(values, numpy.ndarray)
    edges = task.get_output_value("hist_edges")
    assert len(edges) > 0
    assert isinstance(edges[0], numpy.ndarray)



from __future__ import annotations

import logging

import numpy
from silx.gui import qt
from silx.gui.colors import Colormap
from silx.gui.plot.StackView import StackViewMainWindow

from darfix import config
from darfix import dtypes
from darfix.core.dataset import ImageDataset
from darfix.core.noiseremoval import BackgroundType
from darfix.core.noiseremoval import NoiseRemovalOperation
from darfix.core.state_of_operation import Operation

from .operationHistory import OperationHistoryWidget
from .parametersWidget import ParametersWidget

_logger = logging.getLogger(__name__)


class NoiseRemovalWidget(qt.QWidget):
    """
    Widget to apply noise removal from a dataset.
    For now it can apply both background subtraction and hot pixel removal.
    For background subtraction the user can choose the background to use:
    dark frames, low intensity data or all the data. From these background
    frames, an image is computed either using the mean or the median.
    """

    sigLaunchOperation = qt.Signal(dict)

    def __init__(self, parent=None):
        super().__init__(parent)

        self._incomingOperations: list[NoiseRemovalOperation] = []
        self._input_dataset: ImageDataset | None = None
        self.indices: numpy.ndarray | None = None
        self.bg_indices: numpy.ndarray | None = None
        self.bg_dataset: ImageDataset | None = None
        self.setWindowFlags(qt.Qt.WindowType.Widget)

        self._parametersWidget = ParametersWidget()
        self._sv = StackViewMainWindow()
        self._sv.setColormap(
            Colormap(
                name=config.DEFAULT_COLORMAP_NAME,
                normalization=config.DEFAULT_COLORMAP_NORM,
            )
        )
        self._sv.setKeepDataAspectRatio(True)
        self._size = None
        self._method = None
        self._background = self._parametersWidget.bsBackgroundCB.currentText()
        self._bottom_threshold = self._parametersWidget.bottomLE.text()
        self._top_threshold = self._parametersWidget.topLE.text()

        self._operationHistory = OperationHistoryWidget()

        layout = qt.QVBoxLayout()
        layout.addWidget(self._sv)
        layout.addWidget(self._parametersWidget)
        layout.addWidget(self._operationHistory)
        self.setLayout(layout)

        # Add connections
        self._parametersWidget.computeBS.clicked.connect(
            self._launchBackgroundSubtraction
        )
        self._parametersWidget.computeHP.clicked.connect(self._launchHotPixelRemoval)
        self._parametersWidget.computeTP.clicked.connect(self._launchThresholdRemoval)
        self._parametersWidget.computeMR.clicked.connect(self._launchMaskRemoval)

    def refreshPlot(self):
        # TODO is there another way to refresh data plot ?
        self.setStack()

    def setDataset(self, dataset: dtypes.Dataset):
        """Saves the dataset and updates the stack with the dataset data."""

        self._operationHistory.clear()

        self._dataset = dataset.dataset
        self.indices = dataset.indices
        if self._dataset.title != "":
            self._sv.setTitleCallback(lambda idx: self._dataset.title)
        self.setStack()
        self.bg_indices = dataset.bg_indices
        self.bg_dataset = dataset.bg_dataset

        self._parametersWidget.computeBS.show()
        self._parametersWidget.computeHP.show()
        self._parametersWidget.computeTP.show()
        self._parametersWidget.computeMR.show()

        """
        Sets the available background for the user to choose.
        """
        self._parametersWidget.bsBackgroundCB.clear()
        if dataset.bg_dataset is not None:
            self._parametersWidget.bsBackgroundCB.addItem(
                BackgroundType.DARK_DATA.value
            )
        if dataset.bg_indices is not None:
            self._parametersWidget.bsBackgroundCB.addItem(
                BackgroundType.UNUSED_DATA.value
            )
        self._parametersWidget.bsBackgroundCB.addItem(BackgroundType.DATA.value)

    def _launchBackgroundSubtraction(self):
        self._background = self._parametersWidget.bsBackgroundCB.currentText()

        self._method = self._parametersWidget.bsMethodsCB.currentText()

        operation = NoiseRemovalOperation(
            type=Operation.BS,
            parameters={
                "method": self.method,
                "background_type": self._background,
            },
        )
        self._launchOperationInThread(operation)

    def _launchHotPixelRemoval(self):
        self._size = self._parametersWidget.hpSizeCB.currentText()

        operation = NoiseRemovalOperation(
            type=Operation.HP,
            parameters={
                "kernel_size": int(self._size),
            },
        )
        self._launchOperationInThread(operation)

    def _launchThresholdRemoval(self):
        self._bottom_threshold = self._parametersWidget.bottomLE.text()
        self._top_threshold = self._parametersWidget.topLE.text()

        operation = NoiseRemovalOperation(
            type=Operation.THRESHOLD,
            parameters={
                "bottom": int(self._bottom_threshold),
                "top": int(self._top_threshold),
            },
        )
        self._launchOperationInThread(operation)

    def _launchMaskRemoval(self):
        mask = self.mask
        if mask is None:
            return
        operation = NoiseRemovalOperation(
            type=Operation.MASK,
            parameters={"mask": mask},
        )
        self._launchOperationInThread(operation)

    def _launchOperationInThread(self, operation: NoiseRemovalOperation):
        self.sigLaunchOperation.emit(operation)

    def setStack(self, dataset=None):
        """
        Sets new data to the stack.
        Mantains the current frame showed in the view.

        :param Dataset dataset: if not None, data set to the stack will be from the given dataset.
        """
        new_dataset = dataset if dataset is not None else self._dataset
        if new_dataset is None:
            return
        old_nframe = self._sv.getFrameNumber()
        self._sv.setStack(new_dataset.as_array3d(self.indices))
        self._sv.setFrameNumber(old_nframe)

    def setDefaultParameters(self, operations: list[NoiseRemovalOperation]):
        """
        Set default values un widget parameters based on previous user history

        :param history: previous user history
        """
        self._parametersWidget.set_default_values(operations)

    def extendOperationHistory(self):
        self._operationHistory.extend(self._incomingOperations)
        self._incomingOperations.clear()

    def abortIncomingOperations(self):
        for op in self._incomingOperations:
            op["aborted"] = True
        self._operationHistory.extend(self._incomingOperations)
        self._incomingOperations.clear()

    def getOperationHistory(self):
        return self._operationHistory.getOperations()

    def setIncomingOperations(
        self, operations: list[NoiseRemovalOperation], clear=False
    ):
        if self.hasIncomingOperations() and not clear:
            _logger.error(
                f"These operations have not been executed : {self._incomingOperations}"
            )
        self._incomingOperations = operations

    def hasIncomingOperations(self) -> bool:
        return len(self._incomingOperations) > 0

    def getIncomingOperations(self) -> tuple[NoiseRemovalOperation]:
        return self._incomingOperations

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, size):
        self._size = size
        self._parametersWidget.hpSizeCB.setCurrentText(size)

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        self._method = method
        self._parametersWidget.bsMethodsCB.setCurrentText(method)

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, background):
        if self._parametersWidget.bsBackgroundCB.findText(background) >= 0:
            self._background = background
            self._parametersWidget.bsBackgroundCB.setCurrentText(background)

    @property
    def bottom_threshold(self):
        return self._bottom_threshold

    @bottom_threshold.setter
    def bottom_threshold(self, bottom):
        self._bottom_threshold = bottom
        self._parametersWidget.bottomLE.setText(bottom)

    @property
    def top_threshold(self):
        return self._top_threshold

    @top_threshold.setter
    def top_threshold(self, top):
        self._top_threshold = top
        self._parametersWidget.topLE.setText(top)

    @property
    def _dataset(self):
        return self._input_dataset

    @_dataset.setter
    def _dataset(self, dataset):
        self._input_dataset = dataset
        self.__clearMaskWithWrongShape()

    @property
    def mask(self):
        return self._svPlotWidget.getSelectionMask()

    @mask.setter
    def mask(self, mask):
        self.__storeMask(mask)
        self.__clearMaskWithWrongShape()

    def __storeMask(self, mask):
        if mask is None:
            self._svPlotWidget.clearMask()
        else:
            self._svPlotWidget.setSelectionMask(mask)

    @property
    def _svPlotWidget(self):
        return self._sv.getPlotWidget()

    def __clearMaskWithWrongShape(self):
        mask = self.mask
        if mask is None:
            return
        if self._dataset is None or self._dataset.data is None:
            return
        stack_shape = self._dataset.data.shape[-2:]
        mask_shape = mask.shape
        if stack_shape == mask_shape:
            return
        self.__storeMask(mask)



from __future__ import annotations

import os

import numpy
from ewokscore import Task
from ewokscore.missing_data import MISSING_DATA
from ewokscore.missing_data import MissingData
from ewokscore.model import BaseInputModel
from pydantic import ConfigDict
from skimage.transform import rescale

from darfix.dtypes import Dataset


class Inputs(BaseInputModel):
    model_config = ConfigDict(use_attribute_docstrings=True)
    dataset: Dataset
    """ Input dataset containing a stack of images """
    scale: float
    """Factor to rescale images of the dataset."""
    output_dir: str | MissingData = MISSING_DATA
    """ Output directory where the binned data will be saved. If not set, use the input dataset directory."""


class Binning(
    Task,
    input_model=Inputs,
    output_names=["dataset"],
):
    """Rescale images of a Darfix dataset by a given factor."""

    def run(self):
        input_dataset: Dataset = self.inputs.dataset
        dataset = input_dataset.dataset

        if len(dataset.data.shape) >= 4:
            # TODO: Is this expected ? Or should it be fixed for higher dimensionality ?
            raise ValueError("Binning cannot only be applied to 4D datasets or higher")

        scale = self.inputs.scale
        output_dir = self.get_input_value("output_dir", dataset.dir)
        os.makedirs(output_dir, exist_ok=True)

        # rescale data
        new_data = None
        for i, image in enumerate(dataset.data):
            simage = rescale(image, scale, anti_aliasing=True, preserve_range=True)
            if new_data is None:
                new_data = numpy.empty(
                    (len(dataset.data),) + simage.shape, dtype=dataset.data.dtype
                )
            new_data[i] = simage
            if self.cancelled:
                # if cancelled then self.outputs.dataset will be MISSING_DATA
                return

        new_dataset = dataset.copy(new_dir=output_dir, new_data=new_data)

        self.outputs.dataset = Dataset(
            dataset=new_dataset,
            indices=input_dataset.indices,
            bg_indices=input_dataset.bg_indices,
            bg_dataset=input_dataset.bg_dataset,
        )

    def cancel(self) -> None:
        """
        Cancel binning.
        """
        # Cancellation is very simple: binning is done image by image.
        # Between each image binning we check if the task has been cancelled.
        self.cancelled = True



from __future__ import annotations

from silx.gui import qt


def create_completer(filters=None) -> qt.QCompleter:
    completer = qt.QCompleter()
    if qt.BINDING in ("PyQt5", "PySide2"):
        # note: should work with PyQt5 but we get some troubles with it on esrf deployment
        # see https://gitlab.esrf.fr/XRD/darfix/-/issues/174
        model = qt.QDirModel(completer)
    else:
        completer.setCompletionRole(qt.QFileSystemModel.FilePathRole)
        model = qt.QFileSystemModel(completer)
        model.setRootPath(qt.QDir.currentPath())
        model.setOption(qt.QFileSystemModel.DontWatchForChanges)
    if filters is not None:
        model.setFilter(filters)
    completer.setModel(model)

    return completer


def select_output_hdf5_file_with_dialog() -> str | None:
    fileDialog = qt.QFileDialog()

    fileDialog.setFileMode(qt.QFileDialog.FileMode.AnyFile)
    fileDialog.setAcceptMode(qt.QFileDialog.AcceptMode.AcceptSave)
    fileDialog.setOption(qt.QFileDialog.Option.DontUseNativeDialog)
    fileDialog.setDefaultSuffix(".h5")
    if fileDialog.exec():
        return fileDialog.selectedFiles()[0]
    else:
        return None



from __future__ import annotations

from silx.gui import qt

from ...dtypes import AxisType

AXIS_TYPES: dict[str, AxisType] = {
    "motors": "dims",
    "angles (centered)": "center",
    "angles": None,
}


class AxisTypeComboBox(qt.QComboBox):
    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        self.addItems(AXIS_TYPES.keys())

    def getCurrentAxisType(self) -> AxisType:
        axis_type = self.currentText()
        return AXIS_TYPES[axis_type]



from __future__ import annotations

import logging
from typing import Any
from typing import Sequence

import numpy

from darfix.core import array_utils
from darfix.processing import dimension_detection

_logger = logging.getLogger(__file__)


class AcquisitionDims(dict):
    """
    Define the view of the data which has to be made
    """

    def add_dim(self, axis: int, dim: Dimension | dict):
        if isinstance(dim, dict):
            dim = Dimension.from_dict(dim)
        if not isinstance(dim, Dimension):
            raise TypeError(f"dim is expected to be a {Dimension}. Get {type(dim)}")
        self[axis] = dim

    def remove_dim(self, axis: int):
        if axis in self:
            del self[axis]

    @property
    def ndim(self) -> int:
        return len(self)

    def get(self, axis: int, default=None) -> Dimension | None:
        """
        Get Dimension at certain axis.

        :param int axis: axis of the dimension.
        :return: the requested dimension if exists.
        """
        assert type(axis) is int
        return super().get(axis, default)

    def get_names(self) -> list[str]:
        """
        Get list with all the names of the dimensions.

        :return: array_like of strings
        """

        return [dim.name for dim in self.values()]

    @property
    def shape(self) -> tuple[int, ...]:
        """
        Shape order is reversed from the axis so the data is correctly reshaped
        so that the dimensions which motors move first are at the last axis of the
        data. This is done to mantain the axes order as they are used to in the beamlines.

        :return: shape of the currently defined dims
        """
        shape = []
        for iDim in reversed(range(self.ndim)):
            if iDim not in self:
                shape.append(1)
            else:
                shape.append(self[iDim].size or -1)
        return tuple(shape)

    @staticmethod
    def from_dict(raw_dims: dict[int, Any]) -> AcquisitionDims:
        if not isinstance(raw_dims, dict):
            raise TypeError(f"dims should be a dictionary. Got {raw_dims}.")
        dims = AcquisitionDims()
        # sort by key (axis) to sort from fastest to slowest
        for axis in sorted(raw_dims.keys()):
            dims.add_dim(axis, raw_dims[axis])
        return dims

    def to_dict(self) -> dict:
        return {k: v.to_dict() for k, v in self.items()}


class Dimension:
    """
    Define a dimension used during the dataset

    :param str name: name of the dimension (should fit the fabioh5 mapping
                     for now)
    :param Union[int,None] size: length of the dimension.
    """

    def __init__(
        self,
        name: str,
        size: int = 0,
        start: float = 0,
        stop: float = 0,
    ):
        self.__name = name
        self._size = size
        self._start = start
        self._stop = stop

    @property
    def name(self) -> str:
        """
        Name of the dimension (Typically the name of the related positioner)
        """
        return self.__name

    @property
    def size(self) -> int:
        """
        Size of the dimension
        """
        return self._size

    @property
    def start(self) -> float:
        """
        float start value in the linspace of the dimension
        """
        return self._start

    @property
    def stop(self) -> float:
        """
        float stop value in the linspace of the dimension
        """
        return self._stop

    @property
    def step(self) -> float:
        """
        float step in the linspace of the dimension
        """
        if self._size <= 1:
            return 0.0
        return (self._stop - self._start) / (self._size - 1)

    def compute_unique_values(self) -> numpy.ndarray:
        """
        Compute a linspace with dimension parameters

        :returns numpy array: a linspace with dimension parameters

        """
        return numpy.linspace(self.start, self.stop, self._size)

    def __str__(self):
        return f"{self.name} size: {self.size}"

    def guess_parameters(self, values: Sequence[float], tolerance: float):
        """
        Guess size start and stop values of the dimension

        :param array_like values: list of values.

        :param tolerance: Tolerance to find the unique values
        """
        self._size, self._start, self._stop = (
            dimension_detection.find_linspace_parameters(values, tolerance)
        )

    def to_dict(self) -> dict[str, Any]:
        """Translate the current Dimension to a dictionary"""
        return {
            "name": self.name,
            "size": self.size,
            "start": self.start,
            "stop": self.stop,
        }

    @staticmethod
    def from_dict(_dict: dict) -> Dimension:
        """
        This initialize a new dimension instance from a dict.

        Note :

        In darfix 2.x the dimension had a 'range' attribute and since darfix 3.0, 'range' is replaced by 'start and 'stop' attributes.
        This method also ensure old save from a 2.x darfix program can still be readable.

        :param dict _dict: dict defining the dimension. Should contains the
                            following keys: name, size, start and stop.

        :return dimension instance: A new dimension instance
        """

        assert type(_dict) is dict
        if not ("name" in _dict) or not ("size" in _dict):
            raise ValueError(
                "unable to create a valid dim object because 'name' or 'size' is missing"
            )

        if "range" in _dict:
            # darfix 2.x
            if not isinstance(_dict["range"], list) or len(_dict["range"]) != 3:
                raise ValueError(
                    "unable to create a valid dim object because 'range' is not a list of size 3"
                )

            dim = Dimension(
                name=_dict["name"],
                size=_dict["size"],
                start=_dict["range"][0],
                stop=_dict["range"][1],
            )
        else:
            # darfix 3.x
            if not ("start" in _dict) or not ("stop" in _dict):
                raise ValueError(
                    "unable to create a valid dim object because 'start' or 'stop' is missing"
                )
            dim = Dimension(
                name=_dict["name"],
                size=_dict["size"],
                start=_dict["start"],
                stop=_dict["stop"],
            )

        return dim


def find_dimensions_from_metadata(
    metadata: dict[str, numpy.ndarray], tolerance: float
) -> AcquisitionDims:

    dims = AcquisitionDims()

    dimensions = []

    # For every key that has more than one different value, create a new Dimension.
    for key, values in metadata.items():

        dataset_size = len(values)

        unique_values, unique_counts = array_utils.unique(values, return_counts=True)
        if len(unique_values) == 0:
            continue
        if unique_counts[0] == dataset_size:
            continue
        if not numpy.issubdtype(unique_values.dtype, numpy.number):
            continue
        dimension = Dimension(key)
        dimension.guess_parameters(values, tolerance)
        _logger.info(
            "Axis %d: dimension '%s' with %d unique values",
            len(dimensions),
            key,
            len(unique_values),
        )

        # Value that tells when does the change of value occur. It is used to know the order
        # of the reshaping.
        dimension.changing_value = __compute_changing_value(values)
        dimensions.append(dimension)

    for dimension in sorted(dimensions, key=lambda x: x.changing_value, reverse=True):
        dims.add_dim(axis=dims.ndim, dim=dimension)
        _logger.info(
            "Dimension %s of size %d has been added for reshaping",
            dimension.name,
            dimension.size,
        )
    return dims


def __compute_changing_value(values, changing_value=1):
    """
    Recursive method used to calculate how fast is a dimension. The speed of a dimension is the number of
    times the dimension changes of value while moving through the dataset.
    """
    if len(numpy.unique(values)) > 1:
        return __compute_changing_value(
            values[: int(len(values) / 2)], changing_value + 1
        )
    else:
        return changing_value



from __future__ import annotations

from ewokscore.missing_data import MISSING_DATA
from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThread

from darfix.gui.zSumWidget import ZSumWidget
from darfix.tasks.zsum import ZSum


class ZSumWidgetOW(OWEwoksWidgetOneThread, ewokstaskclass=ZSum):
    """
    Widget that compute and display the Z-sum of a dataset
    """

    name = "z sum"
    icon = "icons/zsum.svg"
    want_main_area = True
    want_control_area = False

    _ewoks_inputs_to_hide_from_orange = ("indices", "selected_axis")

    def __init__(self):
        super().__init__()

        self._widget = ZSumWidget(parent=self)
        self.mainArea.layout().addWidget(self._widget)
        # connect signal / slot
        self._widget.sigAxisChanged.connect(self._onAxisChanged)
        self._widget.sigResetFiltering.connect(self._onUncheckFiltering)

    def handleNewSignals(self) -> None:
        dataset = self.get_task_input_value("dataset", None)
        if dataset is None:
            return super().handleNewSignals()

        self._widget.setDataset(dataset)
        self.set_dynamic_input("selected_axis", None)
        self.open()
        return super().handleNewSignals()

    def task_output_changed(self):
        self._widget.setEnabled(True)
        z_sum = self.get_task_output_value("zsum", MISSING_DATA)
        if z_sum is not MISSING_DATA:
            self._widget.setZSum(z_sum)

    def _onAxisChanged(self, selectedAxis: int):
        self.set_default_input("indices", self._widget.indices)
        self.set_dynamic_input("selected_axis", selectedAxis)
        self.execute_ewoks_task_without_propagation()
        self._widget.setDisabled(True)

    def _onUncheckFiltering(self):
        self.set_default_input("indices", None)
        self.set_dynamic_input("selected_axis", None)
        self.execute_ewoks_task_without_propagation()
        self._widget.setDisabled(True)



import numpy
import pytest
from sklearn.datasets import load_iris
from sklearn.decomposition import PCA

from darfix.decomposition.ipca import IPCA
from darfix.decomposition.nica import NICA
from darfix.tests.decomposition.utils import images
from darfix.tests.decomposition.utils import sampler


@pytest.fixture
def X():
    iris = load_iris()
    return iris.data


def test_whiten_pca(X):
    nica = NICA(X, 2)
    pca = PCA(2, whiten=True)
    X = pca.fit_transform(X.T).T
    numpy.testing.assert_equal(X, nica.Z)
    numpy.testing.assert_equal(pca.components_, nica.V)


def test_whiten_ipca(X):
    nica = NICA(X, 2, chunksize=2)
    ipca = IPCA(X, 2, 2, whiten=True, rowvar=False)
    ipca.fit_transform()
    numpy.testing.assert_equal(ipca.W.T, nica.Z)
    numpy.testing.assert_equal(ipca.H, nica.V)


def test_W(X):
    nica = NICA(X, 2)
    nica.fit_transform()
    assert nica.W.shape == (X.shape[0], 2)


def test_H(X):
    nica = NICA(X, 2)
    nica.fit_transform()
    assert nica.H.shape == (2, X.shape[1])


def test_fit_transform(X):
    resources = ["circle", "star", "pentagon", "square"]
    num_images = 100
    means = [15, 30, 45, 60]
    tol = 0.7
    sample = sampler(resources, means)

    X = numpy.array([sample(i).flatten() for i in range(num_images)])
    nica = NICA(X, 4)
    nica.fit_transform(max_iter=1000)

    stack = numpy.asarray(images(resources))
    for img in stack:
        img = img / numpy.linalg.norm(img)
        found = False
        for component in nica.H:
            comp = component.reshape(img.shape)
            comp = comp / numpy.linalg.norm(comp)
            err = numpy.linalg.norm(comp - img)
            if err < tol:
                found = True
                break
        assert found is True


def test_fit_transform_IPCA(X):
    resources = ["circle", "star", "pentagon", "square"]
    num_images = 100
    means = [15, 30, 45, 60]
    tol = 0.7
    sample = sampler(resources, means)

    X = numpy.array([sample(i).flatten() for i in range(num_images)])
    nica = NICA(X, 4, chunksize=1000)
    nica.fit_transform(max_iter=1000)

    stack = numpy.asarray(images(resources))
    for img in stack:
        img = img / numpy.linalg.norm(img)
        found = False
        for component in nica.H:
            comp = component.reshape(img.shape)
            comp = comp / numpy.linalg.norm(comp)
            err = numpy.linalg.norm(comp - img)
            if err < tol:
                found = True
                break
        assert found is True


def test_fit_transform_indices(X):
    resources = ["circle", "star", "pentagon", "square"]
    num_images = 100
    means = [15, 30, 45, 60]
    tol = 0.7
    sample = sampler(resources, means)

    X = numpy.array([sample(i).flatten() for i in range(num_images)])
    nica = NICA(X, 4, indices=numpy.arange(50))
    nica.fit_transform(max_iter=1000)

    stack = numpy.asarray(images(resources[:-1]))
    for img in stack:
        img = img / numpy.linalg.norm(img)
        found = False
        for component in nica.H:
            comp = component.reshape(img.shape)
            comp = comp / numpy.linalg.norm(comp)
            err = numpy.linalg.norm(comp - img)
            if err < tol:
                found = True
                break
        assert found is True


def test_fit_transform_IPCA_indices(X):
    resources = ["circle", "star", "pentagon", "square"]
    num_images = 100
    means = [15, 30, 45, 60]
    tol = 0.7
    sample = sampler(resources, means)

    X = numpy.array([sample(i).flatten() for i in range(num_images)])
    nica = NICA(X, 4, chunksize=1000, indices=numpy.arange(50))
    nica.fit_transform(max_iter=1000)

    stack = numpy.asarray(images(resources[:-1]))
    for img in stack:
        img = img / numpy.linalg.norm(img)
        found = False
        for component in nica.H:
            comp = component.reshape(img.shape)
            comp = comp / numpy.linalg.norm(comp)
            err = numpy.linalg.norm(comp - img)
            if err < tol:
                found = True
                break
        assert found is True



import logging
import multiprocessing as mp
from multiprocessing.connection import Connection

from silx.gui import qt

_logger = logging.getLogger(__name__)


class OperationProcess(qt.QThread):
    """
    A QThread that run an async operation process and emit `finished` signal with the result of the `function` callback as an argument.
    Process can be aborted by calling `kill`. In this case the `finished` argument is None.
    """

    finished = qt.Signal(object)  # emits result or None if aborted

    def __init__(self, parent, function, *args, **kwargs):
        super().__init__(parent)
        self.__parent_conn, self.__child_conn = mp.Pipe(False)
        self.__process = mp.Process(
            target=self._target,
            args=(self.__child_conn, function, args, kwargs),
        )

    def _target(self, conn: Connection, func, args, kwargs):
        """Run function in child process and put result in queue."""
        try:
            res = func(*args, **kwargs)
        except Exception as e:
            res = e
        conn.send(res)

    def run(self):
        self.__process.start()
        self.__process.join()
        result = None
        if self.__parent_conn.poll():
            result = self.__parent_conn.recv()
            if isinstance(result, Exception):
                _logger.error(result, exc_info=True)
                result = None
        self.finished.emit(result)

    def kill(self):
        """Forcefully terminate the process and emit None."""
        self.__process.terminate()



from __future__ import annotations

import logging
import os

import h5py
from ewoksorange.gui.parameterform import block_signals
from silx.gui import qt

from darfix.core.datapathfinder import DETECTOR_KEYWORD
from darfix.core.datapathfinder import SCAN_KEYWORD
from darfix.gui.configuration.action import AdvancedConfigurationAction
from darfix.gui.configuration.action import RequiredConfigurationAction
from darfix.gui.configuration.level import ConfigurationLevel
from darfix.gui.data_selection.hdf5_data_selection_widgets import (
    HDF5DatasetSelectionWidget,
)
from darfix.gui.utils.fileselection import FileSelector
from darfix.tasks.hdf5_scans_concatenation import ConcatenateHDF5Scans
from darfix.tasks.hdf5_scans_concatenation import guess_output_file

_logger = logging.getLogger(__name__)


class HDF5ConcatenateWindow(qt.QMainWindow):
    """Main window giving access to the different options to concatenate HDF5 scans"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        # define toolbar
        toolbar = qt.QToolBar(self)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        self.__configurationModesAction = qt.QAction(self)
        self.__configurationModesAction.setCheckable(False)
        menu = qt.QMenu(self)
        self.__configurationModesAction.setMenu(menu)
        toolbar.addAction(self.__configurationModesAction)

        self.__configurationModesGroup = qt.QActionGroup(self)
        self.__configurationModesGroup.setExclusive(True)
        self.__configurationModesGroup.triggered.connect(self._userModeChanged)

        self._requiredConfigAction = RequiredConfigurationAction(toolbar)
        menu.addAction(self._requiredConfigAction)
        self.__configurationModesGroup.addAction(self._requiredConfigAction)
        self._advancedConfigAction = AdvancedConfigurationAction(toolbar)
        menu.addAction(self._advancedConfigAction)
        self.__configurationModesGroup.addAction(self._advancedConfigAction)

        # define main widget
        self._mainWidget = ConcatenateScansWidget(self)
        self.setCentralWidget(self._mainWidget)

        self._advancedConfigAction.setChecked(True)
        self._userModeChanged(self._advancedConfigAction)

    def getConcatenateScansWidget(self):
        return self._mainWidget

    def _userModeChanged(self, action):
        self.__configurationModesAction.setIcon(action.icon())
        self.__configurationModesAction.setToolTip(action.tooltip())
        if action is self._requiredConfigAction:
            level = ConfigurationLevel.REQUIRED
        elif action is self._advancedConfigAction:
            level = ConfigurationLevel.ADVANCED
        else:
            raise NotImplementedError
        self.setConfigurationLevel(level=level)

    # expose API
    def setConfigurationLevel(self, level: ConfigurationLevel | str):
        self._mainWidget.setConfigurationLevel(level)


class ConcatenateScansWidget(HDF5DatasetSelectionWidget):
    """
    Widget to concatenate a series of scans together.
    Same API as the HDF5DatasetSelectionWidget but adds an instance of HDF5EntriesWidget
    """

    AVAILABLE_KEYWORDS = (
        DETECTOR_KEYWORD,
        SCAN_KEYWORD,
    )

    sigOutputFileChanged = qt.Signal(str)
    sigOverwriteChanged = qt.Signal(bool)
    sigAutoUpdateOutputFile = qt.Signal(bool)
    sigDetectorPathChanged = qt.Signal(str)
    sigMetadataPathChanged = qt.Signal(str)
    sigEntriesToConcatenateChanged = qt.Signal()

    def __init__(
        self,
        parent: qt.QWidget | None = None,
    ) -> None:
        super().__init__(parent)

        # update layout for the exising widget from HDF5DatasetSelectionWidget
        self.layout().addWidget(self._inputFileSelector, 0, 0, 1, 5)
        self.layout().addWidget(self._detectorDataPath, 4, 0, 1, 5)
        self.layout().addWidget(self._positionerDataPath, 5, 0, 1, 5)

        # scan list
        self._scanEntries = HDF5EntriesWidget()
        self._scanEntries.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )
        self.layout().addWidget(self._scanEntries, 1, 0, 2, 5)

        # output file selector
        self._outputFileSelector = FileSelector(label="output file:")
        self._outputFileSelector.setDialogFileMode(qt.QFileDialog.AnyFile)
        self._outputFileSelector.setDialogNameFilters(
            ("HDF5 files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",)
        )
        # overwrite output
        self.layout().addWidget(self._outputFileSelector, 6, 0, 1, 3)
        self._overwriteOutput = qt.QCheckBox("overwrite")
        self._overwriteOutput.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)

        self.layout().addWidget(self._overwriteOutput, 6, 3, 1, 1)

        # auto update
        self._autoUpdate = qt.QCheckBox("auto update")
        self._autoUpdate.setSizePolicy(qt.QSizePolicy.Fixed, qt.QSizePolicy.Fixed)
        self.layout().addWidget(self._autoUpdate, 6, 4, 1, 1)
        self._autoUpdate.setToolTip(
            "update advanced option automatically when input file changed"
        )

        # set up
        self._overwriteOutput.setChecked(True)
        self._autoUpdate.setChecked(True)
        self._detectorDataPath.setPattern(
            ConcatenateHDF5Scans.DEFAULT_DETECTOR_DATA_PATH,
            store_as_default=True,
        )
        self._detectorDataPath.setAvailableKeywords(
            self.AVAILABLE_KEYWORDS,
        )
        self._positionerDataPath.setPattern(
            ConcatenateHDF5Scans.DEFAULT_POSITIONERS_DATA_PATH,
            store_as_default=True,
        )
        self._positionerDataPath.setAvailableKeywords(
            self.AVAILABLE_KEYWORDS,
        )

        # connect signal / slot
        self._outputFileSelector.sigFileChanged.connect(self._outputFileChanged)
        self._positionerDataPath.sigPatternChanged.connect(self.sigMetadataPathChanged)
        self._detectorDataPath.sigPatternChanged.connect(self.sigDetectorPathChanged)
        self._scanEntries.sigSelectedChanged.connect(
            self.sigEntriesToConcatenateChanged
        )

    def _outputFileChanged(self):
        self.sigOutputFileChanged.emit(self._outputFileSelector.getFilePath())

    def _inputFileChanged(self):
        input_file = self.getInputFile()
        if input_file is not None and not os.path.exists(input_file):
            _logger.debug(f"{input_file} doesn't exist")
        else:
            self._updateEntries()
            if self._autoUpdate.isChecked():
                self._deduceOutputFile()
        super()._inputFileChanged()

    # expose API
    def getSelectedEntries(self):
        return self._scanEntries.getSelectedEntries()

    def setSelectedEntries(self, entries: tuple):
        self._scanEntries.setSelectedEntries(entries=entries)

    def _updateEntries(self):
        input_file = self.getInputFile()
        if input_file is None:
            entries = ()
        else:
            with h5py.File(input_file, mode="r") as h5f:
                valid_items = filter(
                    lambda a: isinstance(a, h5py.Group),
                    [h5f.get(item) for item in h5f],
                )
                entries = [item.name for item in valid_items]
        self._scanEntries.setEntries(entries)

    def _deduceOutputFile(self, target_processed_data_dir: bool = True):
        input_file = self.getInputFile()
        if input_file is None:
            return
        if not h5py.is_hdf5(input_file):
            _logger.error("unable to deduce output file. Input is not an HDF5 file")
            return
        output_file = guess_output_file(
            input_file=input_file, target_processed_data_dir=target_processed_data_dir
        )
        self.setOutputFile(output_file)

    def setOutputFile(self, file_path: str):
        self._outputFileSelector.setFilePath(file_path=file_path)
        self.sigOutputFileChanged.emit(file_path)

    def getOutputFile(self) -> str | None:
        return self._outputFileSelector.getFilePath()

    def setOverwrite(self, overwrite: bool):
        self._overwriteOutput.setChecked(overwrite)

    def getOverwrite(self) -> bool:
        return self._overwriteOutput.isChecked()

    def setDetectorPath(self, detector_path: str):
        self._detectorDataPath.setPattern(detector_path)

    def setPositionersGroupPath(self, positioners_group_path: str):
        self._positionerDataPath.setPattern(positioners_group_path)

    def setConfigurationLevel(self, level: ConfigurationLevel | str):
        level = ConfigurationLevel.from_value(level)
        self._positionerDataPath.setVisible(level > ConfigurationLevel.REQUIRED)
        self._detectorDataPath.setVisible(level > ConfigurationLevel.REQUIRED)
        self._outputFileSelector.setVisible(level > ConfigurationLevel.REQUIRED)
        self._overwriteOutput.setVisible(level > ConfigurationLevel.REQUIRED)
        self._autoUpdate.setVisible(level > ConfigurationLevel.REQUIRED)


class HDF5EntriesWidget(qt.QDialog):
    """Dialog to select the entries to use"""

    sigSelectedChanged = qt.Signal()

    def __init__(self) -> None:
        super().__init__()
        self.setLayout(qt.QVBoxLayout())

        # list widget
        self._listWidget = qt.QListWidget()
        self.layout().addWidget(self._listWidget)

        types = qt.QDialogButtonBox.YesToAll | qt.QDialogButtonBox.NoToAll
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        selectAllButton = self._buttons.button(qt.QDialogButtonBox.YesToAll)
        selectAllButton.setText("select all")
        selectAllButton.released.connect(self.selectAll)

        unselectAllButton = self._buttons.button(qt.QDialogButtonBox.NoToAll)
        unselectAllButton.setText("unselect all")
        unselectAllButton.released.connect(self.unselectAll)

    def setEntries(self, entries: tuple[str]):
        """clear existing list, set provided entries and select them"""
        self._listWidget.clear()
        for entry in entries:
            item = qt.QListWidgetItem(parent=self._listWidget)
            checkbox = qt.QCheckBox(entry, parent=self._listWidget)
            checkbox.setChecked(True)
            item.setData(qt.Qt.UserRole, entry)
            self._listWidget.setItemWidget(item, checkbox)
            # connect signal / slot
            checkbox.toggled.connect(self.sigSelectedChanged)

    def _getEntriesItems(self) -> tuple[qt.QListWidgetItem]:
        return [
            self._listWidget.item(i_item) for i_item in range(self._listWidget.count())
        ]

    def getEntries(self) -> tuple[str]:
        return tuple([item.data(qt.Qt.UserRole) for item in self._getEntriesItems()])

    def selectAll(self):
        self.setSelectedEntries(self.getEntries())

    def unselectAll(self):
        self.setSelectedEntries(())

    def getSelectedEntries(self) -> tuple[str]:
        filtered_items = filter(
            lambda a: self._listWidget.itemWidget(a).isChecked(),
            self._getEntriesItems(),
        )
        return tuple([item.data(qt.Qt.UserRole) for item in filtered_items])

    def setSelectedEntries(self, entries: tuple[str]):
        entries = [self._clean_entry_name(entry) for entry in entries]

        with block_signals(self):
            # send sigSelectedChanged once only
            for item in self._getEntriesItems():
                itemWidget = self._listWidget.itemWidget(item)
                checked = self._clean_entry_name(item.data(qt.Qt.UserRole)) in entries
                itemWidget.setChecked(checked)
        self.sigSelectedChanged.emit()

    @staticmethod
    def _clean_entry_name(entry):
        """clean entry names regarding the name use either 1.1 or /1.1"""
        return entry.lstrip("/")



from __future__ import annotations

from silx.gui import qt


class StandardButtonBox(qt.QDialogButtonBox):
    """
    A button box including standard buttons: Ok, Abort and Reset.

    The box has two states that can be changed by using the method `setIsComputing`:
        - setIsComputing(False), default: Ok/Reset enabled. Abort hidden
        - setIsComputing(True): Ok/Reset disabled. Abort visible.
    """

    def __init__(
        self,
        parent,
        additionalButtons: qt.QDialogButtonBox.StandardButton = qt.QDialogButtonBox.StandardButton.NoButton,
    ):
        super().__init__(parent=parent)
        self.setStandardButtons(
            qt.QDialogButtonBox.Ok
            | qt.QDialogButtonBox.Abort
            | qt.QDialogButtonBox.Reset
            | additionalButtons
        )

        self.abortButton.hide()

    @property
    def okButton(self) -> qt.QPushButton:
        okButton = self.button(qt.QDialogButtonBox.Ok)
        assert okButton is not None
        return okButton

    @property
    def resetButton(self) -> qt.QPushButton:
        resetButton = self.button(qt.QDialogButtonBox.Reset)
        assert resetButton is not None
        return resetButton

    @property
    def abortButton(self) -> qt.QPushButton:
        abortButton = self.button(qt.QDialogButtonBox.Abort)
        assert abortButton is not None
        return abortButton

    def setIsComputing(self, isComputing: bool):
        self.okButton.setDisabled(isComputing)
        self.resetButton.setDisabled(isComputing)
        self.abortButton.setVisible(isComputing)



from __future__ import annotations

from typing import Tuple
from typing import Union

import numpy

TwoDVector = Union[numpy.ndarray, Tuple[float, float]]


def apply_2D_ROI(
    img: numpy.ndarray,
    origin: TwoDVector | None = None,
    size: TwoDVector | None = None,
    center: TwoDVector | None = None,
) -> numpy.ndarray:
    """Function that computes a ROI at an image.

    :param array_like img: Image
    :param origin: Origin of the roi
    :param 2d-vector size: [Height, Width] of the roi.
    :param center: Center of the roi
    :raises: AssertionError, ValueError
    """

    assert size is not None, "The size of the roi must be given"

    img = numpy.asanyarray(img)

    if origin is not None:
        assert all(i >= 0 for i in origin) and all(
            j < img.shape[i] for i, j in enumerate(origin)
        ), "Origin must be a valid pixel"
        origin = numpy.array(origin)
        size = numpy.array(size)
        points = numpy.ceil([origin, origin + size]).astype(int)
        points[1] = numpy.minimum(points[1], img.shape)
    elif center is not None:
        assert all(i >= 0 for i in center) and all(
            j < img.shape[i] for i, j in enumerate(center)
        ), "Center must be a valid pixel"
        center = numpy.array(center)
        size = numpy.array(size) * 0.5
        # Compute points and ceil in case of decimal
        points = numpy.ceil([center - size, center + size]).astype(int)
        # Check lower and upper bounds
        points[points < 0] = 0
        points[1] = numpy.minimum(points[1], img.shape)
    else:
        raise ValueError("Origin or center expected")
    return img[points[0, 0] : points[1, 0], points[0, 1] : points[1, 1]]


def apply_3D_ROI(
    data: numpy.ndarray,
    origin: TwoDVector | None = None,
    size: TwoDVector | None = None,
    center: TwoDVector | None = None,
) -> numpy.ndarray:
    """Function that computes the ROI of each image in stack of images.

    :param array_like data: The stack of images
    :param origin: Origin of the roi
    :param 2d-vector size: [Height, Width] of the roi.
    :param center: Center of the roi
    :raises: AssertionError, ValueError
    """
    assert size is not None, "The size of the roi must be given"

    data = numpy.asanyarray(data)

    if origin is not None:
        assert all(i >= 0 for i in origin) and all(
            j < data[0].shape[i] for i, j in enumerate(origin)
        ), "Origin must be a valid pixel"
        origin = numpy.array(origin)
        size = numpy.array(size)
        points = numpy.ceil([origin, origin + size]).astype(int)
        points[1] = numpy.minimum(points[1], data[0].shape)
    elif center is not None:
        assert all(i >= 0 for i in center) and all(
            j < data[0].shape[i] for i, j in enumerate(center)
        ), "Center must be a valid pixel"
        center = numpy.array(center)
        size = numpy.array(size) * 0.5
        # Compute points and ceil in case of decimal
        points = numpy.ceil([center - size, center + size]).astype(int)
        # Check lower and upper bounds
        points[points < 0] = 0
        points[1] = numpy.minimum(points[1], data[0].shape)
    else:
        raise ValueError("Origin or center expected")
    return data[:, points[0, 0] : points[1, 0], points[0, 1] : points[1, 1]]


def clampROI(
    roi_origin: tuple[float, float],
    roi_size: tuple[float, float],
    frame_origin: tuple[float, float],
    frame_size: tuple[int, int],
) -> tuple[tuple[float, float], tuple[int, int]]:
    """
    Clamp given roi to the given frame. Returns origin and size of the clamped roi.

    The 'origin' of both the 'frame' and the 'roi' refers to the locations of the corner with the smallest coordinates.
    """
    roi_origin_x, roi_origin_y = roi_origin
    roi_width, roi_height = roi_size

    frame_origin_x, frame_center_y = frame_origin
    frame_width, frame_height = frame_size

    # Calculate the effective frame boundaries
    frame_min_x = frame_origin_x
    frame_max_x = frame_origin_x + frame_width
    frame_min_y = frame_center_y
    frame_max_y = frame_center_y + frame_height

    # Calculate the effective ROI boundaries
    roi_min_x = roi_origin_x
    roi_max_x = roi_origin_x + roi_width
    roi_min_y = roi_origin_y
    roi_max_y = roi_origin_y + roi_height

    if (frame_min_x > roi_min_x and frame_min_y > roi_min_y) and (
        frame_max_x < roi_max_x and frame_max_y < roi_max_y
    ):
        # if the frame is fully included in the ROI
        # update ROI to match frame shape
        return (frame_origin, frame_size)
    elif (frame_min_x < roi_min_x and frame_min_y < roi_min_y) and (
        frame_max_x > roi_max_x and frame_max_y > roi_max_y
    ):
        # if the roi is fully included in the frame
        return roi_origin, roi_size
    else:
        # in case of intersection
        new_roi_min_x = max(frame_min_x, roi_min_x)
        new_roi_max_x = min(frame_max_x, roi_max_x)
        new_roi_min_y = max(frame_min_y, roi_min_y)
        new_roi_max_y = min(frame_max_y, roi_max_y)

        new_roi_origin = (new_roi_min_x, new_roi_min_y)
        new_roi_size = (
            abs(new_roi_max_x - new_roi_min_x),
            abs(new_roi_max_y - new_roi_min_y),
        )
        return new_roi_origin, new_roi_size
